import unittest
from datetime import datetime, timedelta

from foalorm.orm import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()


class User(db.Entity):
    id = PrimaryKey(int)
    login = Required(str)
    password = Required(str)
    created_at = Required(datetime, default=datetime.now())


class TestLambda(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        setup_database(db)

    @classmethod
    def setUp(self) -> None:
        with db_session:
            User(id=1, login='test', password='123456', created_at=datetime(2012, 12, 13, 5, 25, 30))
            User(id=2, login='test2', password='123456', created_at=datetime(2015, 12, 13, 5, 25, 30))
            User(id=3, login='test3', password='123456')

    @classmethod
    def tearDownClass(cls) -> None:
        teardown_database(db)

    @classmethod
    def tearDown(self) -> None:
        with db_session:
            User.select().delete()

    @db_session
    def test_select(self):
        result = User.select(lambda u: u.created_at < datetime.now() - timedelta(days=365))[:]
        self.assertEqual([u.id for u in result], [1, 2])

    @db_session
    def test_order_by_1(self):
        result = User.select().order_by(lambda u: u.id)
        self.assertEqual([u.id for u in result], [1, 2, 3])

    @db_session
    def test_order_by_2(self):
        result = User.select().order_by(lambda u: desc(u.id))
        self.assertEqual([u.id for u in result], [3, 2, 1])

    @db_session
    def test_order_by_3(self):
        result = User.select().order_by(lambda u: (u.login, u.id))
        self.assertEqual([u.id for u in result], [1, 2, 3])

    @db_session
    def test_order_by_4(self):
        result = User.select().order_by(lambda u: (desc(u.login), desc(u.id)))
        self.assertEqual([u.id for u in result], [3, 2, 1])




from __future__ import absolute_import, print_function, division

import unittest

from foalorm.orm.core import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()

class Department(db.Entity):
    number = PrimaryKey(int)
    groups = Set('Group')

class Group(db.Entity):
    id = PrimaryKey(int)
    dept = Required('Department')
    students = Set('Student')

class Student(db.Entity):
    name = Required(str)
    age = Required(int)
    group = Required('Group')
    scholarship = Required(int, default=0)
    courses = Set('Course')

class Course(db.Entity):
    name = Required(str)
    semester = Required(int)
    PrimaryKey(name, semester)
    students = Set('Student')


class TestDistinct(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            d1 = Department(number=1)
            d2 = Department(number=2)
            g1 = Group(id=1, dept=d1)
            g2 = Group(id=2, dept=d2)
            s1 = Student(id=1, name='S1', age=20, group=g1, scholarship=0)
            s2 = Student(id=2, name='S2', age=21, group=g1, scholarship=100)
            s3 = Student(id=3, name='S3', age=23, group=g1, scholarship=200)
            s4 = Student(id=4, name='S4', age=21, group=g1, scholarship=100)
            s5 = Student(id=5, name='S5', age=23, group=g2, scholarship=0)
            s6 = Student(id=6, name='S6', age=23, group=g2, scholarship=200)
            c1 = Course(name='C1', semester=1, students=[s1, s2, s3])
            c2 = Course(name='C2', semester=1, students=[s2, s3, s5, s6])
            c3 = Course(name='C3', semester=2, students=[s4, s5, s6])

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        db_session.__enter__()

    def tearDown(self):
        db_session.__exit__()

    def test_group_by(self):
        result = set(select((s.age, sum(s.scholarship)) for s in Student if s.scholarship > 0))
        self.assertEqual(result, {(21, 200), (23, 400)})
        self.assertNotIn('distinct', db.last_sql.lower())

    def test_group_by_having(self):
        result = set(select((s.age, sum(s.scholarship)) for s in Student if sum(s.scholarship) < 300))
        self.assertEqual(result, {(20, 0), (21, 200)})
        self.assertNotIn('distinct', db.last_sql.lower())

    def test_aggregation_no_group_by_1(self):
        result = set(select(sum(s.scholarship) for s in Student if s.age < 23))
        self.assertEqual(result, {200})
        self.assertNotIn('distinct', db.last_sql.lower())

    def test_aggregation_no_group_by_2(self):
        result = set(select((sum(s.scholarship), min(s.scholarship)) for s in Student if s.age < 23))
        self.assertEqual(result, {(200, 0)})
        self.assertNotIn('distinct', db.last_sql.lower())

    def test_aggregation_no_group_by_3(self):
        result = set(select((sum(s.scholarship), min(s.scholarship))
                            for s in Student for g in Group
                            if s.group == g and g.dept.number == 1))
        self.assertEqual(result, {(400, 0)})
        self.assertNotIn('distinct', db.last_sql.lower())


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest
from foalorm.orm.core import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()


class Person(db.Entity):
    name = Required(str)
    friends = Set('Person', reverse='friends')


class TestSymmetricM2M(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        with db_session:
            for p in Person.select(): p.delete()
        with db_session:
            db.insert(Person, id=1, name='A')
            db.insert(Person, id=2, name='B')
            db.insert(Person, id=3, name='C')
            db.insert(Person, id=4, name='D')
            db.insert(Person, id=5, name='E')
            db.insert(Person.friends, person=1, person_2=2)
            db.insert(Person.friends, person=2, person_2=1)
            db.insert(Person.friends, person=1, person_2=3)
            db.insert(Person.friends, person=3, person_2=1)
        db_session.__enter__()
    def tearDown(self):
        rollback()
        db_session.__exit__()
    def test1a(self):
        p1 = Person[1]
        p4 = Person[4]
        p1.friends.add(p4)
        self.assertEqual(set(p4.friends), {p1})
    def test1b(self):
        p1 = Person[1]
        p4 = Person[4]
        p1.friends.add(p4)
        self.assertEqual(set(p1.friends), {Person[2], Person[3], p4})
    def test1c(self):
        p1 = Person[1]
        p4 = Person[4]
        p1.friends.add(p4)
        commit()
        rows = db.select("* from person_friends order by person, person_2")
        self.assertEqual(rows, [(1,2), (1,3), (1,4), (2,1), (3,1), (4,1)])
    def test2a(self):
        p1 = Person[1]
        p2 = Person[2]
        p1.friends.remove(p2)
        self.assertEqual(set(p1.friends), {Person[3]})
    def test2b(self):
        p1 = Person[1]
        p2 = Person[2]
        p1.friends.remove(p2)
        self.assertEqual(set(Person[3].friends), {p1})
    def test2c(self):
        p1 = Person[1]
        p2 = Person[2]
        p1.friends.remove(p2)
        self.assertEqual(set(p2.friends), set())
    def test2d(self):
        p1 = Person[1]
        p2 = Person[2]
        p1.friends.remove(p2)
        commit()
        rows = db.select("* from person_friends order by person, person_2")
        self.assertEqual(rows, [(1,3), (3,1)])
    def test3a(self):
        db.execute('delete from person_friends')
        db.insert(Person.friends, person=1, person_2=2)
        p1 = Person[1]
        p2 = Person[2]
        p2_friends = set(p2.friends)
        self.assertEqual(p2_friends, set())
        try: p1_friends = set(p1.friends)
        except UnrepeatableReadError as e: self.assertEqual(e.args[0],
            "Phantom object Person[1] appeared in collection Person[2].friends")
        else: self.fail()
    def test3b(self):
        db.execute('delete from person_friends')
        db.insert(Person.friends, person=1, person_2=2)
        p1 = Person[1]
        p2 = Person[2]
        p1_friends = set(p1.friends)
        self.assertEqual(p1_friends, {p2})
        try: p2_friends = set(p2.friends)
        except UnrepeatableReadError as e: self.assertEqual(e.args[0],
            "Phantom object Person[1] disappeared from collection Person[2].friends")
        else: self.fail()

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest
from foalorm.orm.tests.testutils import raises_exception
from foalorm.orm import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()

class AbstractUser(db.Entity):
    username = PrimaryKey(str)

class User(AbstractUser):
    diagrams = Set('Diagram')
    email = Optional(str)

class SubUser1(User):
    attr1 = Optional(str)

class SubUser2(User):
    attr2 = Optional(str)

class Organization(AbstractUser):
    address = Optional(str)

class SubOrg1(Organization):
    attr3 = Optional(str)

class SubOrg2(Organization):
    attr4 = Optional(str)

class Diagram(db.Entity):
    name = Required(str)
    owner = Required(User)


def is_seed(entity, pk):
    cache = entity._database_._get_cache()
    return pk in [ obj._pk_ for obj in cache.seeds[entity._pk_attrs_] ]


class TestFindInCache(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            u1 = User(username='user1')
            u2 = SubUser1(username='subuser1', attr1='some attr')
            u3 = SubUser2(username='subuser2', attr2='some attr')
            o1 = Organization(username='org1')
            o2 = SubOrg1(username='suborg1', attr3='some attr')
            o3 = SubOrg2(username='suborg2', attr4='some attr')
            au = AbstractUser(username='abstractUser')
            Diagram(name='diagram1', owner=u1)
            Diagram(name='diagram2', owner=u2)
            Diagram(name='diagram3', owner=u3)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test1(self):
        u = User.get(username='org1')
        org = Organization.get(username='org1')
        u1 = User.get(username='org1')
        self.assertEqual(u, None)
        self.assertEqual(org, Organization['org1'])
        self.assertEqual(u1, None)

    def test_user_1(self):
        Diagram.get(lambda d: d.name == 'diagram1')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'user1'))
        u = AbstractUser['user1']
        self.assertNotEqual(last_sql, db.last_sql)
        self.assertEqual(u.__class__, User)

    def test_user_2(self):
        Diagram.get(lambda d: d.name == 'diagram1')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'user1'))
        u = User['user1']
        self.assertNotEqual(last_sql, db.last_sql)
        self.assertEqual(u.__class__, User)

    @raises_exception(ObjectNotFound)
    def test_user_3(self):
        Diagram.get(lambda d: d.name == 'diagram1')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'user1'))
        try:
            SubUser1['user1']
        finally:
            self.assertNotEqual(last_sql, db.last_sql)

    @raises_exception(ObjectNotFound)
    def test_user_4(self):
        Diagram.get(lambda d: d.name == 'diagram1')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'user1'))
        try:
            Organization['user1']
        finally:
            self.assertEqual(last_sql, db.last_sql)

    @raises_exception(ObjectNotFound)
    def test_user_5(self):
        Diagram.get(lambda d: d.name == 'diagram1')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'user1'))
        try:
            SubOrg1['user1']
        finally:
            self.assertEqual(last_sql, db.last_sql)

    def test_subuser_1(self):
        Diagram.get(lambda d: d.name == 'diagram2')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'subuser1'))
        u = AbstractUser['subuser1']
        self.assertNotEqual(last_sql, db.last_sql)
        self.assertEqual(u.__class__, SubUser1)

    def test_subuser_2(self):
        Diagram.get(lambda d: d.name == 'diagram2')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'subuser1'))
        u = User['subuser1']
        self.assertNotEqual(last_sql, db.last_sql)
        self.assertEqual(u.__class__, SubUser1)

    def test_subuser_3(self):
        Diagram.get(lambda d: d.name == 'diagram2')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'subuser1'))
        u = SubUser1['subuser1']
        self.assertNotEqual(last_sql, db.last_sql)
        self.assertEqual(u.__class__, SubUser1)

    @raises_exception(ObjectNotFound)
    def test_subuser_4(self):
        Diagram.get(lambda d: d.name == 'diagram2')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'subuser1'))
        try:
            Organization['subuser1']
        finally:
            self.assertEqual(last_sql, db.last_sql)

    @raises_exception(ObjectNotFound)
    def test_subuser_5(self):
        Diagram.get(lambda d: d.name == 'diagram2')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'subuser1'))
        try:
            SubUser2['subuser1']
        finally:
            self.assertNotEqual(last_sql, db.last_sql)

    @raises_exception(ObjectNotFound)
    def test_subuser_6(self):
        Diagram.get(lambda d: d.name == 'diagram2')
        last_sql = db.last_sql
        self.assertTrue(is_seed(User, 'subuser1'))
        try:
            SubOrg2['subuser1']
        finally:
            self.assertEqual(last_sql, db.last_sql)

    def test_user_6(self):
        u1 = SubUser1['subuser1']
        last_sql = db.last_sql
        u2 = SubUser1['subuser1']
        self.assertEqual(last_sql, db.last_sql)
        self.assertEqual(u1, u2)

    def test_user_7(self):
        u1 = SubUser1['subuser1']
        u1.delete()
        last_sql = db.last_sql
        u2 = SubUser1.get(username='subuser1')
        self.assertEqual(last_sql, db.last_sql)
        self.assertEqual(u2, None)

    def test_user_8(self):
        u1 = SubUser1['subuser1']
        last_sql = db.last_sql
        u2 = SubUser1.get(username='subuser1', attr1='wrong val')
        self.assertEqual(last_sql, db.last_sql)
        self.assertEqual(u2, None)

if __name__ == '__main__':
    unittest.main()


from __future__ import absolute_import, print_function

from bottle import default_app, install, route, request, redirect, run, template

# Import eStore model http://editor.ponyorm.com/user/pony/eStore
from foalorm.orm.examples.estore import *
from foalorm.orm.integration.bottle_plugin import FoalORMPlugin

# After the plugin is installed each request will be processed
# in a separate database session. Once the HTTP request processing
# is finished the plugin does the following:
#  * commit the changes to the database (or rollback if an exception happened)
#  * clear the transaction cache
#  * return the database connection to the connection pool
install(FoalORMPlugin())

@route('/')
@route('/products/')
def all_products():
    # Get the list of all products from the database
    products = select(p for p in Product)
    return template('''
    <h1>List of products</h1>
    <ul>
    %for p in products:
        <li><a href="/products/{{ p.id }}/">{{ p.name }}</a>
    %end
    </ul>
    ''', products=products)

@route('/products/:id/')
def show_product(id):
    # Get the instance of the Product entity by the primary key
    p = Product[id]
    # You can traverse entity relationship attributes inside the template
    # In this examples it is many-to-many relationship p.categories
    # Since the data were not loaded into the cache yet,
    # it will result in a separate SQL query.
    return template('''
    <h1>{{ p.name }}</h1>
    <p>Price: {{ p.price }}</p>
    <p>Product categories:</p>
    <ul>
    %for c in p.categories:
        <li>{{ c.name }}
    %end
    </ul>
    <a href="/products/{{ p.id }}/edit/">Edit product info</a>
    <a href="/products/">Return to all products</a>
    ''', p=p)

@route('/products/:id/edit/')
def edit_product(id):
    # Get the instance of the Product entity and display its attributes
    p = Product[id]
    return template('''
    <form action='/products/{{ p.id }}/edit/' method='post'>
      <table>
        <tr>
          <td>Product name:</td>
          <td><input type="text" name="name" value="{{ p.name }}">
        </tr>
        <tr>
          <td>Product price:</td>
          <td><input type="text" name="price" value="{{ p.price }}">
        </tr>
      </table>
      <input type="submit" value="Save!">
    </form>
    <p><a href="/products/{{ p.id }}/">Discard changes</a>
    <p><a href="/products/">Return to all products</a>
    ''', p=p)

@route('/products/:id/edit/', method='POST')
def save_product(id):
    # Get the instance of the Product entity
    p = Product[id]
    # Update the attributes with the new values
    p.name = request.forms.get('name')
    p.price = request.forms.get('price')
    # We might put the commit() command here, but it is not necessary
    # because FoalORMPlugin will take care of this.
    redirect("/products/%d/" % p.id)
    # The Bottle's redirect function raises the HTTPResponse exception.
    # Normally FoalORMPlugin closes the session with rollback
    # if a callback function raises an exception. But in this case
    # FoalORMPlugin understands that this exception is not the error
    # and closes the session with commit.


run(debug=True, host='localhost', port=8080, reloader=True)



from __future__ import absolute_import, print_function, division
from foalorm.py23compat import int_types

from operator import attrgetter

from foalorm.orm import core
from foalorm.orm.core import log_sql, DBSchemaError, MappingError
from foalorm.utils import throw

class DBSchema(object):
    dialect = None
    inline_fk_syntax = True
    named_foreign_keys = True
    def __init__(schema, provider, uppercase=True):
        schema.provider = provider
        schema.tables = {}
        schema.constraints = {}
        schema.indent = '  '
        schema.command_separator = ';\n\n'
        schema.uppercase = uppercase
        schema.names = {}
    def column_list(schema, columns):
        quote_name = schema.provider.quote_name
        return '(%s)' % ', '.join(quote_name(column.name) for column in columns)
    def case(schema, s):
        if schema.uppercase: return s.upper().replace('%S', '%s') \
            .replace(')S', ')s').replace('%R', '%r').replace(')R', ')r')
        else: return s.lower()
    def add_table(schema, table_name, entity=None):
        return schema.table_class(table_name, schema, entity)
    def order_tables_to_create(schema):
        tables = []
        created_tables = set()
        split = schema.provider.split_table_name
        tables_to_create = sorted(schema.tables.values(), key=lambda table: split(table.name))
        while tables_to_create:
            for table in tables_to_create:
                if table.parent_tables.issubset(created_tables):
                    created_tables.add(table)
                    tables_to_create.remove(table)
                    break
            else: table = tables_to_create.pop()
            tables.append(table)
        return tables
    def generate_create_script(schema):
        created_tables = set()
        commands = []
        for table in schema.order_tables_to_create():
            for db_object in table.get_objects_to_create(created_tables):
                commands.append(db_object.get_create_command())
        return schema.command_separator.join(commands)
    def create_tables(schema, provider, connection):
        created_tables = set()
        for table in schema.order_tables_to_create():
            for db_object in table.get_objects_to_create(created_tables):
                base_name = provider.base_name(db_object.name)
                name = db_object.exists(provider, connection, case_sensitive=False)
                if name is None: db_object.create(provider, connection)
                elif name != base_name:
                    quote_name = schema.provider.quote_name
                    n1, n2 = quote_name(db_object.name), quote_name(name)
                    tn1, tn2 = db_object.typename, db_object.typename.lower()
                    throw(DBSchemaError, '%s %s cannot be created, because %s %s ' \
                                         '(with a different letter case) already exists in the database. ' \
                                         'Try to delete %s %s first.' % (tn1, n1, tn2, n2, n2, tn2))
    def check_tables(schema, provider, connection):
        cursor = connection.cursor()
        split = provider.split_table_name
        for table in sorted(schema.tables.values(), key=lambda table: split(table.name)):
            alias = provider.base_name(table.name)
            sql_ast = [ 'SELECT',
                        [ 'ALL', ] + [ [ 'COLUMN', alias, column.name ] for column in table.column_list ],
                        [ 'FROM', [ alias, 'TABLE', table.name ] ],
                        [ 'WHERE', [ 'EQ', [ 'VALUE', 0 ], [ 'VALUE', 1 ] ] ]
                      ]
            sql, adapter = provider.ast2sql(sql_ast)
            if core.local.debug: log_sql(sql)
            provider.execute(cursor, sql)

class DBObject(object):
    def create(table, provider, connection):
        sql = table.get_create_command()
        if core.local.debug: log_sql(sql)
        cursor = connection.cursor()
        provider.execute(cursor, sql)

class Table(DBObject):
    typename = 'Table'
    def __init__(table, name, schema, entity=None):
        if name in schema.tables:
            throw(DBSchemaError, "Table %r already exists in database schema" % name)
        if name in schema.names:
            throw(DBSchemaError, "Table %r cannot be created, name is already in use" % name)
        schema.tables[name] = table
        schema.names[name] = table
        table.schema = schema
        table.name = name
        table.column_list = []
        table.column_dict = {}
        table.indexes = {}
        table.pk_index = None
        table.foreign_keys = {}
        table.parent_tables = set()
        table.child_tables = set()
        table.entities = set()
        table.options = {}
        if entity is not None:
            table.entities.add(entity)
            table.options = entity._table_options_
        table.m2m = set()
    def __repr__(table):
        return '<Table(%s)>' % table.schema.provider.format_table_name(table.name)
    def add_entity(table, entity):
        for e in table.entities:
            if e._root_ is not entity._root_:
                throw(MappingError, "Entities %s and %s cannot be mapped to table %s "
                                   "because they don't belong to the same hierarchy"
                                   % (e, entity, table.name))
        assert '_table_options_' not in entity.__dict__
        table.entities.add(entity)
    def exists(table, provider, connection, case_sensitive=True):
        return provider.table_exists(connection, table.name, case_sensitive)
    def get_create_command(table):
        schema = table.schema
        case = schema.case
        provider = schema.provider
        quote_name = provider.quote_name
        if_not_exists = False # provider.table_if_not_exists_syntax and provider.index_if_not_exists_syntax
        cmd = []
        if not if_not_exists: cmd.append(case('CREATE TABLE %s (') % quote_name(table.name))
        else: cmd.append(case('CREATE TABLE IF NOT EXISTS %s (') % quote_name(table.name))
        for column in table.column_list:
            cmd.append(schema.indent + column.get_sql() + ',')
        if len(table.pk_index.columns) > 1:
            cmd.append(schema.indent + table.pk_index.get_sql() + ',')
        indexes = [ index for index in table.indexes.values()
                    if not index.is_pk and index.is_unique and len(index.columns) > 1 ]
        for index in indexes: assert index.name is not None
        indexes.sort(key=attrgetter('name'))
        for index in indexes: cmd.append(schema.indent+index.get_sql() + ',')
        if not schema.named_foreign_keys:
            for foreign_key in sorted(table.foreign_keys.values(), key=lambda fk: fk.name):
                if schema.inline_fk_syntax and len(foreign_key.child_columns) == 1: continue
                cmd.append(schema.indent+foreign_key.get_sql() + ',')
        interleave_fks = [ fk for fk in table.foreign_keys.values() if fk.interleave ]
        if interleave_fks:
            assert len(interleave_fks) == 1
            fk = interleave_fks[0]
            cmd.append(schema.indent+fk.get_sql())
            cmd.append(case(') INTERLEAVE IN PARENT %s (%s)') % (
                quote_name(fk.parent_table.name),
                ', '.join(quote_name(col.name) for col in fk.child_columns)
            ))
        else:
            cmd[-1] = cmd[-1][:-1]
            cmd.append(')')
        for name, value in sorted(table.options.items()):
            option = table.format_option(name, value)
            if option: cmd.append(option)
        return '\n'.join(cmd)
    def format_option(table, name, value):
        if value is True:
            return name
        if value is False:
            return None
        return '%s %s' % (name, value)
    def get_objects_to_create(table, created_tables=None):
        if created_tables is None: created_tables = set()
        created_tables.add(table)
        result = [ table ]
        indexes = [ index for index in table.indexes.values() if not index.is_pk and not index.is_unique ]
        for index in indexes: assert index.name is not None
        indexes.sort(key=attrgetter('name'))
        result.extend(indexes)
        schema = table.schema
        if schema.named_foreign_keys:
            for foreign_key in sorted(table.foreign_keys.values(), key=lambda fk: fk.name):
                if foreign_key.parent_table not in created_tables: continue
                result.append(foreign_key)
            for child_table in table.child_tables:
                if child_table not in created_tables: continue
                for foreign_key in sorted(child_table.foreign_keys.values(), key=lambda fk: fk.name):
                    if foreign_key.parent_table is not table: continue
                    result.append(foreign_key)
        return result
    def add_column(table, column_name, sql_type, converter, is_not_null=None, sql_default=None):
        return table.schema.column_class(column_name, table, sql_type, converter, is_not_null, sql_default)
    def add_index(table, index_name, columns, is_pk=False, is_unique=None, m2m=False):
        assert index_name is not False
        if index_name is True: index_name = None
        if index_name is None and not is_pk:
            provider = table.schema.provider
            index_name = provider.get_default_index_name(table.name, (column.name for column in columns),
                                                         is_pk=is_pk, is_unique=is_unique, m2m=m2m)
        index = table.indexes.get(columns)
        if index and index.name == index_name and index.is_pk == is_pk and index.is_unique == is_unique:
            return index
        return table.schema.index_class(index_name, table, columns, is_pk, is_unique)
    def add_foreign_key(table, fk_name, child_columns, parent_table, parent_columns, index_name=None, on_delete=False,
                        interleave=False):
        if fk_name is None:
            provider = table.schema.provider
            child_column_names = tuple(column.name for column in child_columns)
            fk_name = provider.get_default_fk_name(table.name, parent_table.name, child_column_names)
        return table.schema.fk_class(fk_name, table, child_columns, parent_table, parent_columns, index_name, on_delete,
                                     interleave=interleave)

class Column(object):
    auto_template = '%(type)s PRIMARY KEY AUTOINCREMENT'
    def __init__(column, name, table, sql_type, converter, is_not_null=None, sql_default=None):
        if name in table.column_dict:
            throw(DBSchemaError, "Column %r already exists in table %r" % (name, table.name))
        table.column_dict[name] = column
        table.column_list.append(column)
        column.table = table
        column.name = name
        column.sql_type = sql_type
        column.converter = converter
        column.is_not_null = is_not_null
        column.sql_default = sql_default
        column.is_pk = False
        column.is_pk_part = False
        column.is_unique = False
    def __repr__(column):
        return '<Column(%s.%s)>' % (column.table.name, column.name)
    def get_sql(column):
        table = column.table
        schema = table.schema
        quote_name = schema.provider.quote_name
        case = schema.case
        result = []
        append = result.append
        append(quote_name(column.name))

        def add_default():
            if column.sql_default not in (None, True, False):
                append(case('DEFAULT'))
                append(column.sql_default)

        if column.is_pk == 'auto' and column.auto_template and column.converter.py_type in int_types:
            append(case(column.auto_template % dict(type=column.sql_type)))
            add_default()
        else:
            append(case(column.sql_type))
            add_default()
            if column.is_pk:
                if schema.dialect == 'SQLite': append(case('NOT NULL'))
                append(case('PRIMARY KEY'))
            else:
                if column.is_unique: append(case('UNIQUE'))
                if column.is_not_null: append(case('NOT NULL'))
        if schema.inline_fk_syntax and not schema.named_foreign_keys:
            foreign_key = table.foreign_keys.get((column,))
            if foreign_key is not None:
                parent_table = foreign_key.parent_table
                append(case('REFERENCES'))
                append(quote_name(parent_table.name))
                append(schema.column_list(foreign_key.parent_columns))
                if foreign_key.on_delete:
                    append('ON DELETE %s' % foreign_key.on_delete)
        return ' '.join(result)

class Constraint(DBObject):
    def __init__(constraint, name, schema):
        if name is not None:
            assert name not in schema.names
            if name in schema.constraints: throw(DBSchemaError,
                "Constraint with name %r already exists" % name)
            schema.names[name] = constraint
            schema.constraints[name] = constraint
        constraint.schema = schema
        constraint.name = name

class DBIndex(Constraint):
    typename = 'Index'
    def __init__(index, name, table, columns, is_pk=False, is_unique=None):
        assert len(columns) > 0
        for column in columns:
            if column.table is not table: throw(DBSchemaError,
                "Column %r does not belong to table %r and cannot be part of its index"
                % (column.name, table.name))
        if columns in table.indexes:
            if len(columns) == 1: throw(DBSchemaError, "Index for column %r already exists" % columns[0].name)
            else: throw(DBSchemaError, "Index for columns (%s) already exists" % ', '.join(repr(column.name) for column in columns))
        if is_pk:
            if table.pk_index is not None: throw(DBSchemaError,
                'Primary key for table %r is already defined' % table.name)
            table.pk_index = index
            if is_unique is None: is_unique = True
            elif not is_unique: throw(DBSchemaError,
                "Incompatible combination of is_unique=False and is_pk=True")
        elif is_unique is None: is_unique = False
        schema = table.schema
        if name is not None and name in schema.names:
            throw(DBSchemaError, 'Index %s cannot be created, name is already in use' % name)
        Constraint.__init__(index, name, schema)
        for column in columns:
            column.is_pk = column.is_pk or (len(columns) == 1 and is_pk)
            column.is_pk_part = column.is_pk_part or bool(is_pk)
            column.is_unique = column.is_unique or (is_unique and len(columns) == 1)
        table.indexes[columns] = index
        index.table = table
        index.columns = columns
        index.is_pk = is_pk
        index.is_unique = is_unique
    def exists(index, provider, connection, case_sensitive=True):
        return provider.index_exists(connection, index.table.name, index.name, case_sensitive)
    def get_sql(index):
        return index._get_create_sql(inside_table=True)
    def get_create_command(index):
        return index._get_create_sql(inside_table=False)
    def _get_create_sql(index, inside_table):
        schema = index.schema
        case = schema.case
        quote_name = schema.provider.quote_name
        cmd = []
        append = cmd.append
        if not inside_table:
            if index.is_pk: throw(DBSchemaError,
                'Primary key index cannot be defined outside of table definition')
            append(case('CREATE'))
            if index.is_unique: append(case('UNIQUE'))
            append(case('INDEX'))
            # if schema.provider.index_if_not_exists_syntax:
            #     append(case('IF NOT EXISTS'))
            append(quote_name(index.name))
            append(case('ON'))
            append(quote_name(index.table.name))
            converter = index.columns[0].converter
            if isinstance(converter.py_type, core.Array) and converter.provider.dialect == 'PostgreSQL':
                append(case('USING GIN'))
        else:
            if index.name:
                append(case('CONSTRAINT'))
                append(quote_name(index.name))
            if index.is_pk: append(case('PRIMARY KEY'))
            elif index.is_unique: append(case('UNIQUE'))
            else: append(case('INDEX'))
        append(schema.column_list(index.columns))
        return ' '.join(cmd)

class ForeignKey(Constraint):
    typename = 'Foreign key'
    def __init__(foreign_key, name, child_table, child_columns, parent_table, parent_columns, index_name, on_delete,
                 interleave=False):
        schema = parent_table.schema
        if schema is not child_table.schema: throw(DBSchemaError,
            'Parent and child tables of foreign_key cannot belong to different schemata')
        for column in parent_columns:
            if column.table is not parent_table: throw(DBSchemaError,
                'Column %r does not belong to table %r' % (column.name, parent_table.name))
        for column in child_columns:
            if column.table is not child_table: throw(DBSchemaError,
                'Column %r does not belong to table %r' % (column.name, child_table.name))
        if len(parent_columns) != len(child_columns): throw(DBSchemaError,
            'Foreign key columns count do not match')
        if child_columns in child_table.foreign_keys:
            if len(child_columns) == 1: throw(DBSchemaError, 'Foreign key for column %r already defined' % child_columns[0].name)
            else: throw(DBSchemaError, 'Foreign key for columns (%s) already defined' % ', '.join(repr(column.name) for column in child_columns))
        if name is not None and name in schema.names:
            throw(DBSchemaError, 'Foreign key %s cannot be created, name is already in use' % name)
        Constraint.__init__(foreign_key, name, schema)
        child_table.foreign_keys[child_columns] = foreign_key
        if child_table is not parent_table:
            child_table.parent_tables.add(parent_table)
            parent_table.child_tables.add(child_table)
        foreign_key.parent_table = parent_table
        foreign_key.parent_columns = parent_columns
        foreign_key.child_table = child_table
        foreign_key.child_columns = child_columns
        foreign_key.on_delete = on_delete
        foreign_key.interleave = interleave

        if index_name is not False:
            child_columns_len = len(child_columns)
            if all(columns[:child_columns_len] != child_columns for columns in child_table.indexes):
                child_table.add_index(index_name, child_columns, is_pk=False,
                                      is_unique=False, m2m=bool(child_table.m2m))

    def exists(foreign_key, provider, connection, case_sensitive=True):
        return provider.fk_exists(connection, foreign_key.child_table.name, foreign_key.name, case_sensitive)
    def get_sql(foreign_key):
        return foreign_key._get_create_sql(inside_table=True)
    def get_create_command(foreign_key):
        return foreign_key._get_create_sql(inside_table=False)
    def _get_create_sql(foreign_key, inside_table):
        schema = foreign_key.schema
        case = schema.case
        quote_name = schema.provider.quote_name
        cmd = []
        append = cmd.append
        if not inside_table:
            append(case('ALTER TABLE'))
            append(quote_name(foreign_key.child_table.name))
            append(case('ADD'))
        if schema.named_foreign_keys and foreign_key.name:
            append(case('CONSTRAINT'))
            append(quote_name(foreign_key.name))
        append(case('FOREIGN KEY'))
        append(schema.column_list(foreign_key.child_columns))
        append(case('REFERENCES'))
        append(quote_name(foreign_key.parent_table.name))
        append(schema.column_list(foreign_key.parent_columns))
        if foreign_key.on_delete:
            append(case('ON DELETE %s' % foreign_key.on_delete))
        return ' '.join(cmd)

DBSchema.table_class = Table
DBSchema.column_class = Column
DBSchema.index_class = DBIndex
DBSchema.fk_class = ForeignKey



import sys
import os
import logging

from ponytest import with_cli_args, pony_fixtures, provider_validators, provider, Fixture, \
        ValidationError

from functools import wraps, partial
import click
from contextlib import contextmanager, closing

from foalorm.utils import cached_property, class_property

from contextlib import contextmanager, ContextDecorator

import unittest

from foalorm.orm import db_session, Database, rollback, delete

from io import StringIO

from multiprocessing import Process

import threading

class DBContext(ContextDecorator):

    fixture = 'db'
    enabled = False

    def __init__(self, Test):
        if not isinstance(Test, type):
            # FIXME ?
            TestCls = type(Test)
            NewClass = type(TestCls.__name__, (TestCls,), {})
            NewClass.__module__ = TestCls.__module__
            NewClass.db = property(lambda t: self.db)
            Test.__class__ = NewClass
        else:
            Test.db = class_property(lambda cls: self.db)
        self.Test = Test

    @class_property
    def fixture_name(cls):
        return cls.db_provider

    @class_property
    def db_provider(cls):
        # is used in tests
        return cls.provider_key

    def init_db(self):
        raise NotImplementedError

    @cached_property
    def db(self):
        raise NotImplementedError

    def __enter__(self):
        self.init_db()
        try:
            self.Test.make_entities()
        except (AttributeError, TypeError):
            # No method make_entities with due signature
            pass
        else:
            self.db.generate_mapping(check_tables=True, create_tables=True)
        return self.db

    def __exit__(self, *exc_info):
        self.db.provider.disconnect()

    @classmethod
    def validate_fixtures(cls, fixtures, config):
        return any(f.fixture_key == 'db' for f in fixtures)

    db_name = 'testdb'


@provider()
class GenerateMapping(ContextDecorator):

    weight = 200
    fixture = 'generate_mapping'

    def __init__(self, Test):
        self.Test = Test

    def __enter__(self):
        db = getattr(self.Test, 'db', None)
        if not db or not db.entities:
            return
        for entity in db.entities.values():
            if entity._database_.schema is None:
                db.generate_mapping(check_tables=True, create_tables=True)
            break

    def __exit__(self, *exc_info):
        pass

@provider()
class MySqlContext(DBContext):
    provider_key  = 'mysql'

    def drop_db(self, cursor):
        cursor.execute('use sys')
        cursor.execute('drop database %s' % self.db_name)


    def init_db(self):
        from foalorm.orm.dbproviders.mysql import mysql_module
        with closing(mysql_module.connect(**self.CONN).cursor()) as c:
            try:
                self.drop_db(c)
            except mysql_module.DatabaseError as exc:
                print('Failed to drop db: %s' % exc)
            c.execute('create database %s' % self.db_name)
            c.execute('use %s' % self.db_name)

    CONN = {
        'host': "localhost",
        'user': "ponytest",
        'passwd': "ponytest",
    }

    @cached_property
    def db(self):
        CONN = dict(self.CONN, db=self.db_name)
        return Database('mysql', **CONN)

@provider()
class SqlServerContext(DBContext):

    provider_key = 'sqlserver'

    def get_conn_string(self, db=None):
        s = (
            'DSN=MSSQLdb;'
            'SERVER=mssql;'
            'UID=sa;'
            'PWD=pass;'
        )
        if db:
            s += 'DATABASE=%s' % db
        return s

    @cached_property
    def db(self):
        CONN = self.get_conn_string(self.db_name)
        return Database('mssqlserver', CONN)

    def init_db(self):
        import pyodbc
        cursor = pyodbc.connect(self.get_conn_string(), autocommit=True).cursor()
        with closing(cursor) as c:
            try:
                self.drop_db(c)
            except pyodbc.DatabaseError as exc:
                print('Failed to drop db: %s' % exc)
            c.execute('''CREATE DATABASE %s DEFAULT CHARACTER SET utf8 DEFAULT COLLATE utf8_general_ci''' % self.db_name )
            c.execute('use %s' % self.db_name)

    def drop_db(self, cursor):
        cursor.execute('use master')
        cursor.execute('drop database %s' % self.db_name)


class SqliteMixin(DBContext):

    def init_db(self):
        try:
            os.remove(self.db_path)
        except OSError as exc:
            print('Failed to drop db: %s' % exc)

    @cached_property
    def db_path(self):
        p = os.path.dirname(__file__)
        p = os.path.join(p, '%s.sqlite' % self.db_name)
        return os.path.abspath(p)

    @cached_property
    def db(self):
        return Database('sqlite', self.db_path, create_db=True)


@provider()
class SqliteNoJson1(SqliteMixin):
    provider_key = 'sqlite_no_json1'
    enabled = True

    def __init__(self, cls):
        self.Test = cls
        cls.no_json1 = True
        return super(SqliteNoJson1, self).__init__(cls)

    def __enter__(self):
        resource = super(SqliteNoJson1, self).__enter__()
        self.json1_available = self.Test.db.provider.json1_available
        self.Test.db.provider.json1_available = False
        return resource

    def __exit__(self, *exc_info):
        self.Test.db.provider.json1_available = self.json1_available
        return super(SqliteNoJson1, self).__exit__(*exc_info)


@provider()
class SqliteJson1(SqliteMixin):
    provider_key = 'sqlite_json1'

    def __enter__(self):
        result = super(SqliteJson1, self).__enter__()
        if not self.db.provider.json1_available:
            raise unittest.SkipTest
        return result


@provider()
class PostgresContext(DBContext):
    provider_key = 'postgresql'

    def get_conn_dict(self, no_db=False):
        d = dict(
            user='ponytest', password='ponytest',
            host='localhost', database='postgres',
        )
        if not no_db:
            d.update(database=self.db_name)
        return d

    def init_db(self):
        import psycopg2
        conn = psycopg2.connect(
            **self.get_conn_dict(no_db=True)
        )
        conn.set_isolation_level(0)
        with closing(conn.cursor()) as cursor:
            try:
                self.drop_db(cursor)
            except psycopg2.DatabaseError as exc:
                print('Failed to drop db: %s' % exc)
            cursor.execute('create database %s' % self.db_name)

    def drop_db(self, cursor):
        cursor.execute('drop database %s' % self.db_name)


    @cached_property
    def db(self):
        return Database('postgres', **self.get_conn_dict())


@provider()
class OracleContext(DBContext):
    provider_key = 'oracle'

    def __enter__(self):
        os.environ.update(dict(
            ORACLE_BASE='/u01/app/oracle',
            ORACLE_HOME='/u01/app/oracle/product/12.1.0/dbhome_1',
            ORACLE_OWNR='oracle',
            ORACLE_SID='orcl',
        ))
        return super(OracleContext, self).__enter__()

    def init_db(self):

        import cx_Oracle
        with closing(self.connect_sys()) as conn:
            with closing(conn.cursor()) as cursor:
                try:
                    self._destroy_test_user(cursor)
                except cx_Oracle.DatabaseError as exc:
                    print('Failed to drop user: %s' % exc)
                try:
                    self._drop_tablespace(cursor)
                except cx_Oracle.DatabaseError as exc:
                    print('Failed to drop db: %s' % exc)
                cursor.execute(
                """CREATE TABLESPACE %(tblspace)s
                DATAFILE '%(datafile)s' SIZE 20M
                REUSE AUTOEXTEND ON NEXT 10M MAXSIZE %(maxsize)s
                """ % self.parameters)
                cursor.execute(
                """CREATE TEMPORARY TABLESPACE %(tblspace_temp)s
                TEMPFILE '%(datafile_tmp)s' SIZE 20M
                REUSE AUTOEXTEND ON NEXT 10M MAXSIZE %(maxsize_tmp)s
                """ % self.parameters)
                self._create_test_user(cursor)


    def _drop_tablespace(self, cursor):
        cursor.execute(
            'DROP TABLESPACE %(tblspace)s INCLUDING CONTENTS AND DATAFILES CASCADE CONSTRAINTS'
        % self.parameters)
        cursor.execute(
            'DROP TABLESPACE %(tblspace_temp)s INCLUDING CONTENTS AND DATAFILES CASCADE CONSTRAINTS'
        % self.parameters)


    parameters = {
        'tblspace': 'test_tblspace',
        'tblspace_temp': 'test_tblspace_temp',
        'datafile': 'test_datafile.dbf',
        'datafile_tmp': 'test_datafile_tmp.dbf',
        'user': 'ponytest',
        'password': 'ponytest',
        'maxsize': '100M',
        'maxsize_tmp': '100M',
    }

    def connect_sys(self):
        import cx_Oracle
        return cx_Oracle.connect('sys/the@localhost/ORCL', mode=cx_Oracle.SYSDBA)

    def connect_test(self):
        import cx_Oracle
        return cx_Oracle.connect('ponytest/ponytest@localhost/ORCL')


    @cached_property
    def db(self):
        return Database('oracle', 'ponytest/ponytest@localhost/ORCL')

    def _create_test_user(self, cursor):
        cursor.execute(
        """CREATE USER %(user)s
            IDENTIFIED BY %(password)s
            DEFAULT TABLESPACE %(tblspace)s
            TEMPORARY TABLESPACE %(tblspace_temp)s
            QUOTA UNLIMITED ON %(tblspace)s
        """ % self.parameters
        )
        cursor.execute(
        """GRANT CREATE SESSION,
                    CREATE TABLE,
                    CREATE SEQUENCE,
                    CREATE PROCEDURE,
                    CREATE TRIGGER
            TO %(user)s
        """ % self.parameters
        )

    def _destroy_test_user(self, cursor):
        cursor.execute('''
            DROP USER %(user)s CASCADE
        ''' % self.parameters)


@provider(fixture='log', weight=100, enabled=False)
@contextmanager
def logging_context(test):
    level = logging.getLogger().level
    from foalorm.orm.core import debug, sql_debug
    logging.getLogger().setLevel(logging.INFO)
    sql_debug(True)
    yield
    logging.getLogger().setLevel(level)
    sql_debug(debug)

@provider(fixture='log_all', weight=-100, enabled=False)
def log_all(Test):
    return logging_context(Test)


# @with_cli_args
# @click.option('--log', 'scope', flag_value='test')
# @click.option('--log-all', 'scope', flag_value='all')
# def use_logging(scope):
#     if scope == 'test':
#         yield logging_context
#     elif scope =='all':
#         yield log_all


# @provider(enabled=False)
# class DBSessionProvider(object):
#
#     fixture= 'db_session'
#
#     weight = 30
#
#     def __new__(cls, test):
#         return db_session


@provider(fixture='rollback', weight=40)
@contextmanager
def do_rollback(test):
    try:
        yield
    finally:
        rollback()


@provider()
class SeparateProcess(object):

    # TODO read failures from sep process better

    fixture = 'separate_process'
    enabled = False

    def __init__(self, Test):
        self.Test = Test

    def __call__(self, func):
        def wrapper(Test):
            rnr = unittest.runner.TextTestRunner()
            TestCls = Test if isinstance(Test, type) else type(Test)
            def runTest(self):
                try:
                    func(Test)
                finally:
                    rnr.stream = unittest.runner._WritelnDecorator(StringIO())
            name = getattr(func, '__name__', 'runTest')
            Case = type(TestCls.__name__, (TestCls,), {name: runTest})
            Case.__module__ = TestCls.__module__
            case = Case(name)
            suite = unittest.suite.TestSuite([case])
            def run():
                result = rnr.run(suite)
                if not result.wasSuccessful():
                    sys.exit(1)
            p = Process(target=run, args=())
            p.start()
            p.join()
            case.assertEqual(p.exitcode, 0)
        return wrapper

    @classmethod
    def validate_chain(cls, fixtures, klass):
        for f in fixtures:
            if f.KEY in ('ipdb', 'ipdb_all'):
                return False
        for f in fixtures:
            if f.KEY == 'db' and f.provider_key in ('sqlserver', 'oracle'):
                return True

@provider()
class ClearTables(ContextDecorator):

    fixture = 'clear_tables'

    def __init__(self, test):
        self.test = test

    def __enter__(self):
        pass

    @db_session
    def __exit__(self, *exc_info):
        db = self.test.db
        for entity in db.entities.values():
            if entity._database_.schema is None:
                break
            delete(i for i in entity)


import signal

@provider()
class Timeout(object):

    fixture = 'timeout'

    @with_cli_args
    @click.option('--timeout', type=int)
    def __init__(self, Test, timeout):
        self.Test = Test
        self.timeout = timeout if timeout else Test.TIMEOUT

    enabled = False

    class Exception(Exception):
        pass

    class FailedInSubprocess(Exception):
        pass

    def __call__(self, func):
        def wrapper(test):
            p = Process(target=func, args=(test,))
            p.start()

            def on_expired():
                p.terminate()

            t = threading.Timer(self.timeout, on_expired)
            t.start()
            p.join()
            t.cancel()
            if p.exitcode == -signal.SIGTERM:
                raise self.Exception
            elif p.exitcode:
                raise self.FailedInSubprocess

        return wrapper

    @classmethod
    @with_cli_args
    @click.option('--timeout', type=int)
    def validate_chain(cls, fixtures, klass, timeout):
        if not getattr(klass, 'TIMEOUT', None) and not timeout:
            return False
        for f in fixtures:
            if f.KEY in ('ipdb', 'ipdb_all'):
                return False
        for f in fixtures:
            if f.KEY == 'db' and f.provider_key in ('sqlserver', 'oracle'):
                return True


pony_fixtures['test'].extend([
    'log',
    'clear_tables',
])

pony_fixtures['class'].extend([
    'separate_process',
    'timeout',
    'db',
    'log_all',
    'generate_mapping',
])



from __future__ import absolute_import, print_function, division

import unittest
from datetime import date, datetime
from hashlib import md5

from foalorm.orm.tests.testutils import raises_exception
from foalorm.orm import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()


class User(db.Entity):
    name = Required(str)
    password = Required(str)
    created_at = Required(datetime)

    def __init__(self, name, password):
        password = md5(password.encode('utf8')).hexdigest()
        super(User, self).__init__(name=name, password=password, created_at=datetime.now())
        self.uppercase_name = name.upper()


class TestCustomInit(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(self):
        teardown_database(db)

    def test1(self):
        with db_session:
            u1 = User('John', '123')
            u2 = User('Mike', '456')
            commit()
            self.assertEqual(u1.name, 'John')
            self.assertEqual(u1.uppercase_name, 'JOHN')
            self.assertEqual(u1.password, md5(b'123').hexdigest())
            self.assertEqual(u2.name, 'Mike')
            self.assertEqual(u2.uppercase_name, 'MIKE')
            self.assertEqual(u2.password, md5(b'456').hexdigest())

        with db_session:
            users = select(u for u in User).order_by(User.id)
            self.assertEqual(len(users), 2)
            u1, u2 = users
            self.assertEqual(u1.name, 'John')
            self.assertTrue(not hasattr(u1, 'uppercase_name'))
            self.assertEqual(u1.password, md5(b'123').hexdigest())
            self.assertEqual(u2.name, 'Mike')
            self.assertTrue(not hasattr(u2, 'uppercase_name'))
            self.assertEqual(u2.password, md5(b'456').hexdigest())

if __name__ == '__main__':
    unittest.main()


import io, sys, unittest
import contextlib
from decimal import Decimal
from datetime import date

from foalorm.orm import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()


class Student(db.Entity):
    name = Required(str)
    scholarship = Optional(int)
    gpa = Optional(Decimal, 3, 1)
    dob = Optional(date)
    group = Required('Group')
    courses = Set('Course')
    biography = Optional(LongUnicode)

class Group(db.Entity):
    number = PrimaryKey(int)
    students = Set(Student)

class Course(db.Entity):
    name = Required(str, unique=True)
    students = Set(Student)


normal_stdout = sys.stdout


class TestShow(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

        with db_session:
            g1 = Group(number=1)
            g2 = Group(number=2)
            c1 = Course(name='Math')
            c2 = Course(name='Physics')
            c3 = Course(name='Computer Science')
            Student(id=1, name='S1', group=g1, gpa=3.1, courses=[c1, c2], biography='some text')
            Student(id=2, name='S2', group=g1, gpa=3.2, scholarship=100, dob=date(2000, 1, 1))
            Student(id=3, name='S3', group=g1, gpa=3.3, scholarship=200, dob=date(2001, 1, 2), courses=[c2, c3])

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_1(self):
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            Student.select().show()
        self.assertEqual('\n' + f.getvalue().replace(' ', '~'), '''
id|name|scholarship|gpa|dob~~~~~~~|group~~~
--+----+-----------+---+----------+--------
1~|S1~~|None~~~~~~~|3.1|None~~~~~~|Group[1]
2~|S2~~|100~~~~~~~~|3.2|2000-01-01|Group[1]
3~|S3~~|200~~~~~~~~|3.3|2001-01-02|Group[1]
''')

    def test_2(self):
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            Group.select().show()
        self.assertEqual('\n' + f.getvalue().replace(' ', '~'), '''
number
------
1~~~~~
2~~~~~
''')


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest
from datetime import date, datetime, timedelta

from foalorm.orm.core import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()


class Entity1(db.Entity):
    id = PrimaryKey(int)
    d = Required(date)
    dt = Required(datetime)


class TestDate(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            Entity1(id=1, d=date(2009, 10, 20), dt=datetime(2009, 10, 20, 10, 20, 30))
            Entity1(id=2, d=date(2010, 10, 21), dt=datetime(2010, 10, 21, 10, 21, 31))
            Entity1(id=3, d=date(2011, 11, 22), dt=datetime(2011, 11, 22, 10, 20, 32))

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_create(self):
        e1 = Entity1(id=4, d=date(2011, 10, 20), dt=datetime(2009, 10, 20, 10, 20, 30))

    def test_date_year(self):
        result = select(e for e in Entity1 if e.d.year > 2009)
        self.assertEqual(len(result), 2)

    def test_date_month(self):
        result = select(e for e in Entity1 if e.d.month == 10)
        self.assertEqual(len(result), 2)

    def test_date_day(self):
        result = select(e for e in Entity1 if e.d.day == 22)
        self.assertEqual(len(result), 1)

    def test_datetime_year(self):
        result = select(e for e in Entity1 if e.dt.year > 2009)
        self.assertEqual(len(result), 2)

    def test_datetime_month(self):
        result = select(e for e in Entity1 if e.dt.month == 10)
        self.assertEqual(len(result), 2)

    def test_datetime_day(self):
        result = select(e for e in Entity1 if e.dt.day == 22)
        self.assertEqual(len(result), 1)

    def test_datetime_hour(self):
        result = select(e for e in Entity1 if e.dt.hour == 10)
        self.assertEqual(len(result), 3)

    def test_datetime_minute(self):
        result = select(e for e in Entity1 if e.dt.minute == 20)
        self.assertEqual(len(result), 2)

    def test_datetime_second(self):
        result = select(e for e in Entity1 if e.dt.second == 30)
        self.assertEqual(len(result), 1)

    def test_date_sub_date(self):
        dt = date(2012, 1, 1)
        result = select(e.id for e in Entity1 if dt - e.d > timedelta(days=500))
        self.assertEqual(set(result), {1})

    def test_datetime_sub_datetime(self):
        dt = datetime(2012, 1, 1, 10, 20, 30)
        result = select(e.id for e in Entity1 if dt - e.dt > timedelta(days=500))
        self.assertEqual(set(result), {1})

    def test_date_sub_timedelta_param(self):
        td = timedelta(days=500)
        result = select(e.id for e in Entity1 if e.d - td < date(2009, 1, 1))
        self.assertEqual(set(result), {1})

    def test_date_sub_const_timedelta(self):
        result = select(e.id for e in Entity1 if e.d - timedelta(days=500) < date(2009, 1, 1))
        self.assertEqual(set(result), {1})

    def test_datetime_sub_timedelta_param(self):
        td = timedelta(days=500)
        result = select(e.id for e in Entity1 if e.dt - td < datetime(2009, 1, 1, 10, 20, 30))
        self.assertEqual(set(result), {1})

    def test_datetime_sub_const_timedelta(self):
        result = select(e.id for e in Entity1 if e.dt - timedelta(days=500) < datetime(2009, 1, 1, 10, 20, 30))
        self.assertEqual(set(result), {1})

    def test_date_add_timedelta_param(self):
        td = timedelta(days=500)
        result = select(e.id for e in Entity1 if e.d + td > date(2013, 1, 1))
        self.assertEqual(set(result), {3})

    def test_date_add_const_timedelta(self):
        result = select(e.id for e in Entity1 if e.d + timedelta(days=500) > date(2013, 1, 1))
        self.assertEqual(set(result), {3})

    def test_datetime_add_timedelta_param(self):
        td = timedelta(days=500)
        result = select(e.id for e in Entity1 if e.dt + td > date(2013, 1, 1))
        self.assertEqual(set(result), {3})

    def test_datetime_add_const_timedelta(self):
        result = select(e.id for e in Entity1 if e.dt + timedelta(days=500) > date(2013, 1, 1))
        self.assertEqual(set(result), {3})

    @raises_exception(TypeError, "Unsupported operand types 'date' and 'str' for operation '-' in expression: e.d - s")
    def test_date_sub_error(self):
        s = 'hello'
        result = select(e.id for e in Entity1 if e.d - s > timedelta(days=500))
        self.assertEqual(set(result), {1})

    @raises_exception(TypeError,
                      "Unsupported operand types 'datetime' and 'str' for operation '-' in expression: e.dt - s")
    def test_datetime_sub_error(self):
        s = 'hello'
        result = select(e.id for e in Entity1 if e.dt - s > timedelta(days=500))
        self.assertEqual(set(result), {1})


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest

from foalorm.orm.core import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()

db = Database('sqlite', ':memory:')

class Product(db.Entity):
    id = PrimaryKey(int)
    name = Required(str)
    comments = Set('Comment')

    @property
    def sum_01(self):
        return coalesce(select(c.points for c in self.comments).sum(), 0)

    @property
    def sum_02(self):
        return coalesce(select(c.points for c in self.comments).sum(), 0.0)

    @property
    def sum_03(self):
        return coalesce(select(sum(c.points) for c in self.comments), 0)

    @property
    def sum_04(self):
        return coalesce(select(sum(c.points) for c in self.comments), 0.0)

    @property
    def sum_05(self):
        return sum(c.points for c in self.comments)

    @property
    def sum_06(self):
        return coalesce(sum(c.points for c in self.comments), 0)

    @property
    def sum_07(self):
        return coalesce(sum(c.points for c in self.comments), 0.0)

    @property
    def sum_08(self):
        return select(sum(c.points) for c in self.comments)

    @property
    def sum_09(self):
        return select(coalesce(sum(c.points), 0) for c in self.comments)

    @property
    def sum_10(self):
        return select(coalesce(sum(c.points), 0.0) for c in self.comments)

    @property
    def sum_11(self):
        return select(sum(c.points) for c in self.comments)

    @property
    def sum_12(self):
        return sum(self.comments.points)

    @property
    def sum_13(self):
        return coalesce(sum(self.comments.points), 0)

    @property
    def sum_14(self):
        return coalesce(sum(self.comments.points), 0.0)


class Comment(db.Entity):
    id = PrimaryKey(int)
    points = Required(int)
    product = Optional('Product')


class TestQuerySetMonad(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            p1 = Product(id=1, name='P1')
            p2 = Product(id=2, name='P1', comments=[
                Comment(id=201, points=5)
            ])
            p3 = Product(id=3, name='P1', comments=[
                Comment(id=301, points=1), Comment(id=302, points=2)
            ])
            p4 = Product(id=4, name='P1', comments=[
                Comment(id=401, points=1), Comment(id=402, points=5), Comment(id=403, points=1)
            ])

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_sum_01(self):
        q = list(Product.select().sort_by(lambda p: p.sum_01))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_02(self):
        q = list(Product.select().sort_by(lambda p: p.sum_02))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_03(self):
        q = list(Product.select().sort_by(lambda p: p.sum_03))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_04(self):
        q = list(Product.select().sort_by(lambda p: p.sum_04))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_05(self):
        q = list(Product.select().sort_by(lambda p: p.sum_05))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_06(self):
        q = list(Product.select().sort_by(lambda p: p.sum_06))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_07(self):
        q = list(Product.select().sort_by(lambda p: p.sum_07))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_08(self):
        q = list(Product.select().sort_by(lambda p: p.sum_08))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_09(self):
        q = list(Product.select().sort_by(lambda p: p.sum_09))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_10(self):
        q = list(Product.select().sort_by(lambda p: p.sum_10))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_11(self):
        q = list(Product.select().sort_by(lambda p: p.sum_11))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_12(self):
        q = list(Product.select().sort_by(lambda p: p.sum_12))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_13(self):
        q = list(Product.select().sort_by(lambda p: p.sum_13))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])

    def test_sum_14(self):
        q = list(Product.select().sort_by(lambda p: p.sum_14))
        result = [p.id for p in q]
        self.assertEqual(result, [1, 3, 2, 4])


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest

from foalorm.orm.core import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()

class Group(db.Entity):
    id = PrimaryKey(int)
    students = Set('Student')

class Student(db.Entity):
    name = Required(str)
    age = Required(int)
    group = Required('Group')
    scholarship = Required(int, default=0)
    courses = Set('Course')

class Course(db.Entity):
    name = Required(str)
    semester = Required(int)
    PrimaryKey(name, semester)
    students = Set('Student')


class TestQuerySetMonad(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            g1 = Group(id=1)
            g2 = Group(id=2)
            s1 = Student(id=1, name='S1', age=20, group=g1, scholarship=0)
            s2 = Student(id=2, name='S2', age=23, group=g1, scholarship=100)
            s3 = Student(id=3, name='S3', age=23, group=g2, scholarship=500)
            c1 = Course(name='C1', semester=1, students=[s1, s2])
            c2 = Course(name='C2', semester=1, students=[s2, s3])
            c3 = Course(name='C3', semester=2, students=[s3])

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_len(self):
        result = set(select(g for g in Group if len(g.students) > 1))
        self.assertEqual(result, {Group[1]})

    def test_len_2(self):
        result = set(select(g for g in Group if len(s for s in Student if s.group == g) > 1))
        self.assertEqual(result, {Group[1]})

    def test_len_3(self):
        result = set(select(g for g in Group if len(s.name for s in Student if s.group == g) > 1))
        self.assertEqual(result, {Group[1]})

    def test_count_1(self):
        result = set(select(g for g in Group if count(s.name for s in g.students) > 1))
        self.assertEqual(result, {Group[1]})

    def test_count_2(self):
        result = set(select(g for g in Group if select(s.name for s in g.students).count() > 1))
        self.assertEqual(result, {Group[1]})

    def test_count_3(self):
        result = set(select(s for s in Student if count(c for c in s.courses) > 1))
        self.assertEqual(result, {Student[2], Student[3]})

    def test_count_3a(self):
        result = set(select(s for s in Student if select(c for c in s.courses).count() > 1))
        self.assertEqual(result, {Student[2], Student[3]})
        self.assertTrue('DISTINCT' in db.last_sql)

    def test_count_3b(self):
        result = set(select(s for s in Student if select(c for c in s.courses).count(distinct=False) > 1))
        self.assertEqual(result, {Student[2], Student[3]})
        self.assertTrue('DISTINCT' not in db.last_sql)

    def test_count_4(self):
        result = set(select(c for c in Course if count(s for s in c.students) > 1))
        self.assertEqual(result, {Course['C1', 1], Course['C2', 1]})

    def test_count_5(self):
        result = select(c.semester for c in Course).count(distinct=True)
        self.assertEqual(result, 2)

    def test_count_6(self):
        result = select(c for c in Course).count()
        self.assertEqual(result, 3)
        self.assertTrue('DISTINCT' not in db.last_sql)

    def test_count_7(self):
        result = select(c for c in Course).count(distinct=True)
        self.assertEqual(result, 3)
        self.assertTrue('DISTINCT' in db.last_sql)

    def test_count_8(self):
        select(count(c.semester, distinct=False) for c in Course)[:]
        self.assertTrue('DISTINCT' not in db.last_sql)

    @raises_exception(TypeError, "`distinct` value should be True or False. Got: s.name.startswith('P')")
    def test_count_9(self):
        select(count(s, distinct=s.name.startswith('P')) for s in Student)

    def test_count_10(self):
        select(count('*', distinct=True) for s in Student)[:]
        self.assertTrue('DISTINCT' not in db.last_sql)

    @raises_exception(TypeError)
    def test_sum_1(self):
        result = set(select(g for g in Group if sum(s for s in Student if s.group == g) > 1))

    @raises_exception(TypeError)
    def test_sum_2(self):
        select(g for g in Group if sum(s.name for s in Student if s.group == g) > 1)

    def test_sum_3(self):
        result = sum(s.scholarship for s in Student)
        self.assertEqual(result, 600)

    def test_sum_4(self):
        result = sum(s.scholarship for s in Student if s.name == 'Unnamed')
        self.assertEqual(result, 0)

    def test_sum_5(self):
        result = select(c.semester for c in Course).sum()
        self.assertEqual(result, 4)

    def test_sum_6(self):
        result = select(c.semester for c in Course).sum(distinct=True)
        self.assertEqual(result, 3)

    def test_sum_7(self):
        result = set(select(g for g in Group if sum(s.scholarship for s in Student if s.group == g) > 500))
        self.assertEqual(result, set())

    def test_sum_8(self):
        result = set(select(g for g in Group if select(s.scholarship for s in g.students).sum() > 200))
        self.assertEqual(result, {Group[2]})
        self.assertTrue('DISTINCT' not in db.last_sql)

    def test_sum_9(self):
        result = set(select(g for g in Group if select(s.scholarship for s in g.students).sum(distinct=True) > 200))
        self.assertEqual(result, {Group[2]})
        self.assertTrue('DISTINCT' in db.last_sql)

    def test_sum_10(self):
        select(sum(s.scholarship, distinct=True) for s in Student)[:]
        self.assertTrue('SUM(DISTINCT' in db.last_sql)

    def test_min_1(self):
        result = set(select(g for g in Group if min(s.name for s in Student if s.group == g) == 'S1'))
        self.assertEqual(result, {Group[1]})

    @raises_exception(TypeError)
    def test_min_2(self):
        select(g for g in Group if min(s for s in Student if s.group == g) == None)

    def test_min_3(self):
        result = set(select(g for g in Group if select(s.scholarship for s in g.students).min() == 0))
        self.assertEqual(result, {Group[1]})

    def test_min_4(self):
        result = select(s.scholarship for s in Student).min()
        self.assertEqual(0, result)

    def test_max_1(self):
        result = set(select(g for g in Group if max(s.scholarship for s in Student if s.group == g) > 100))
        self.assertEqual(result, {Group[2]})

    @raises_exception(TypeError)
    def test_max_2(self):
        select(g for g in Group if max(s for s in Student if s.group == g) == None)

    def test_max_3(self):
        result = set(select(g for g in Group if select(s.scholarship for s in g.students).max() == 100))
        self.assertEqual(result, {Group[1]})

    def test_max_4(self):
        result = select(s.scholarship for s in Student).max()
        self.assertEqual(result, 500)

    def test_avg_1(self):
        result = select(g for g in Group if avg(s.scholarship for s in Student if s.group == g) == 50)[:]
        self.assertEqual(result, [Group[1]])

    def test_avg_2(self):
        result = set(select(g for g in Group if select(s.scholarship for s in g.students).avg() == 50))
        self.assertEqual(result, {Group[1]})

    def test_avg_3(self):
        result = select(c.semester for c in Course).avg()
        self.assertAlmostEqual(1.33, result, places=2)

    def test_avg_4(self):
        result = select(c.semester for c in Course).avg(distinct=True)
        self.assertAlmostEqual(1.5, result)

    def test_avg_5(self):
        result = set(select(g for g in Group if select(s.scholarship for s in g.students).avg(distinct=True) == 50))
        self.assertEqual(result, {Group[1]})
        self.assertTrue('AVG(DISTINCT' in db.last_sql)

    def test_avg_6(self):
        select(avg(s.scholarship, distinct=True) for s in Student)[:]
        self.assertTrue('AVG(DISTINCT' in db.last_sql)

    def test_exists_1(self):
        result = set(select(g for g in Group if exists(s for s in g.students if s.age < 23)))
        self.assertEqual(result, {Group[1]})

    def test_exists_2(self):
        result = set(select(g for g in Group if exists(s.age < 23 for s in g.students)))
        self.assertEqual(result, {Group[1]})

    def test_exists_3(self):
        result = set(select(g for g in Group if (s.age < 23 for s in g.students)))
        self.assertEqual(result, {Group[1]})

    def test_negate(self):
        result = set(select(g for g in Group if not(s.scholarship for s in Student if s.group == g)))
        self.assertEqual(result, set())

    def test_no_conditions(self):
        students = set(select(s for s in Student if s.group in (g for g in Group)))
        self.assertEqual(students, {Student[1], Student[2], Student[3]})

    def test_no_conditions_2(self):
        students = set(select(s for s in Student if s.scholarship == max(s.scholarship for s in Student)))
        self.assertEqual(students, {Student[3]})

    def test_hint_join_1(self):
        result = set(select(s for s in Student if JOIN(s.group in select(g for g in Group if g.id < 2))))
        self.assertEqual(result, {Student[1], Student[2]})

    def test_hint_join_2(self):
        result = set(select(s for s in Student if JOIN(s.group not in select(g for g in Group if g.id < 2))))
        self.assertEqual(result, {Student[3]})

    def test_hint_join_3(self):
        result = set(select(s for s in Student if JOIN(s.scholarship in
                        select(s.scholarship + 100 for s in Student if s.name != 'S2'))))
        self.assertEqual(result, {Student[2]})

    def test_hint_join_4(self):
        result = set(select(g for g in Group if JOIN(g in select(s.group for s in g.students))))
        self.assertEqual(result, {Group[1], Group[2]})

    def test_group_concat_1(self):
        result = select(s.name for s in Student).group_concat()
        self.assertEqual(result, 'S1,S2,S3')

    def test_group_concat_2(self):
        result = select(s.name for s in Student).group_concat('-')
        self.assertEqual(result, 'S1-S2-S3')

    def test_group_concat_3(self):
        result = select(s for s in Student if s.name in group_concat(s.name for s in Student))[:]
        self.assertEqual(set(result), {Student[1], Student[2], Student[3]})

    def test_group_concat_4(self):
        result = Student.select().group_concat()
        self.assertEqual(result, '1,2,3')

    def test_group_concat_5(self):
        result = Student.select().group_concat('.')
        self.assertEqual(result, '1.2.3')

    @raises_exception(TypeError, '`group_concat` cannot be used with entity with composite primary key')
    def test_group_concat_6(self):
        select(group_concat(s.courses, '-') for s in Student)

    def test_group_concat_7(self):
        result = select(group_concat(c.semester) for c in Course)[:]
        self.assertEqual(result[0], '1,1,2')

    def test_group_concat_8(self):
        result = select(group_concat(c.semester, '-') for c in Course)[:]
        self.assertEqual(result[0], '1-1-2')

    def test_group_concat_9(self):
        result = select(group_concat(c.semester, distinct=True) for c in Course)[:]
        self.assertEqual(result[0], '1,2')

    def test_group_concat_10(self):
        result = group_concat((s.name for s in Student if int(s.name[1]) > 1), sep='-')
        self.assertEqual(result, 'S2-S3')

    def test_group_concat_11(self):
        result = group_concat((c.semester for c in Course), distinct=True)
        self.assertEqual(result, '1,2')


    @raises_exception(TypeError, 'Query can only iterate over entity or another query (not a list of objects)')
    def test_select_from_select_1(self):
        query = select(s for s in Student if s.scholarship > 0)[:]
        result = set(select(x for x in query))
        self.assertEqual(result, {})

    def test_select_from_select_2(self):
        p, q = 50, 400
        query = select(s for s in Student if s.scholarship > p)
        result = select(x.id for x in query if x.scholarship < q)[:]
        self.assertEqual(set(result), {2})

    def test_select_from_select_3(self):
        p, q = 50, 400
        g = (s for s in Student if s.scholarship > p)
        result = select(x.id for x in g if x.scholarship < q)[:]
        self.assertEqual(set(result), {2})

    def test_select_from_select_4(self):
        p, q = 50, 400
        result = select(x.id for x in (s for s in Student if s.scholarship > p)
                             if x.scholarship < q)[:]
        self.assertEqual(set(result), {2})

    def test_select_from_select_5(self):
        p, q = 50, 400
        result = select(x.id for x in select(s for s in Student if s.scholarship > 0)
                             if x.scholarship < 400)[:]
        self.assertEqual(set(result), {2})

    def test_select_from_select_6(self):
        query = select(s.name for s in Student if s.scholarship > 0)
        result = select(x for x in query if not x.endswith('3'))
        self.assertEqual(set(result), {'S2'})

    @raises_exception(TranslationError, 'Too many values to unpack "for a, b in select(s for ...)" (expected 2, got 1)')
    def test_select_from_select_7(self):
        query = select(s for s in Student if s.scholarship > 0)
        result = select(a for a, b in query)

    @raises_exception(NotImplementedError, 'Please unpack a tuple of (s.name, s.group) in for-loop '
                                           'to individual variables (like: "for x, y in ...")')
    def test_select_from_select_8(self):
        query = select((s.name, s.group) for s in Student if s.scholarship > 0)
        result = select(x for x in query)

    @raises_exception(TranslationError, 'Not enough values to unpack "for x, y in '
                                        'select(s.name, s.group, s.scholarship for ...)" (expected 2, got 3)')
    def test_select_from_select_9(self):
        query = select((s.name, s.group, s.scholarship) for s in Student if s.scholarship > 0)
        result = select(x for x, y in query)

    def test_select_from_select_10(self):
        query = select((s.name, s.age) for s in Student if s.scholarship > 0)
        result = select(n for n, a in query if n.endswith('2') and a > 20)
        self.assertEqual(set(x for x in result), {'S2'})

    def test_aggregations_1(self):
        query = select((min(s.age), max(s.scholarship)) for s in Student)
        result = query[:]
        self.assertEqual(result, [(20, 500)])

    def test_aggregations_2(self):
        query = select((min(s.age), max(s.scholarship)) for s in Student for g in Group)
        result = query[:]
        self.assertEqual(result, [(20, 500)])


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest
from foalorm.orm.core import *
from foalorm.orm.tests import db_params, teardown_database

db = Database()


class Group(db.Entity):
    number = PrimaryKey(int)
    subjects = Set("Subject")


class Subject(db.Entity):
    name = PrimaryKey(str)
    groups = Set(Group)


class TestManyToManyNonComposite(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        db.bind(**db_params)
        db.generate_mapping(check_tables=False)
        db.drop_all_tables(with_all_data=True)

    def setUp(self):
        db.create_tables()
        with db_session:
           g1 = Group(number=101)
           g2 = Group(number=102)
           s1 = Subject(name='Subj1')
           s2 = Subject(name='Subj2')
           s3 = Subject(name='Subj3')
           s4 = Subject(name='Subj4')
           g1.subjects = [ s1, s2 ]

    def tearDown(self):
        teardown_database(db)

    def test_1(self):
        schema = db.schema
        m2m_table_name = 'Group_Subject'
        if not (db.provider.dialect == 'SQLite' and foalorm.__version__ < '0.9'):
            m2m_table_name = m2m_table_name.lower()
        self.assertIn(m2m_table_name, schema.tables)
        m2m_table = schema.tables[m2m_table_name]
        if foalorm.__version__ >= '0.9':
            fkeys = m2m_table.foreign_keys
        else:
            fkeys = set(m2m_table.foreign_keys.values())
        self.assertEqual(len(fkeys), 2)
        for fk in fkeys:
            self.assertEqual(fk.on_delete, 'CASCADE')

    def test_2(self):
        with db_session:
            g = Group.get(number=101)
            s = Subject.get(name='Subj1')
            g.subjects.add(s)

        with db_session:
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    def test_3(self):
        with db_session:
            g = Group.get(number=101)
            s = Subject.get(name='Subj3')
            g.subjects.add(s)

        with db_session:
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2', 'Subj3'})

    def test_4(self):
        with db_session:
            g = Group.get(number=101)
            s = Subject.get(name='Subj3')
            g.subjects.remove(s)

        with db_session:
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    def test_5(self):
        with db_session:
            g = Group.get(number=101)
            s = Subject.get(name='Subj2')
            g.subjects.remove(s)

        with db_session:
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1'})

    def test_6(self):
        with db_session:
            g = Group.get(number=101)
            s1, s2, s3, s4 = Subject.select()[:]
            g.subjects.remove([s1, s2])
            g.subjects.add([s3, s4])

        with db_session:
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj3', 'Subj4'})
            self.assertEqual(Group[101].subjects, {Subject['Subj3'], Subject['Subj4']})

    def test_7(self):
        with db_session:
            g = Group.get(number=101)
            s = Subject.get(name='Subj3')
            g.subjects.add(s)
            g.subjects.remove(s)
            last_sql = db.last_sql

        with db_session:
            self.assertEqual(db.last_sql, last_sql)  # assert no DELETE statement on commit
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    def test_8(self):
        with db_session:
            g = Group.get(number=101)
            s = Subject.get(name='Subj1')
            g.subjects.remove(s)
            g.subjects.add(s)
            last_sql = db.last_sql

        with db_session:
            self.assertEqual(db.last_sql, last_sql)  # assert no INSERT statement on commit
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    def test_9(self):
        with db_session:
            g = Group.get(number=101)
            s1 = Subject.get(name='Subj1')
            s2 = Subject.get(name='Subj2')
            g.subjects.clear()
            g.subjects.add([s1, s2])
            last_sql = db.last_sql

        with db_session:
            self.assertEqual(db.last_sql, last_sql)  # assert no INSERT statement on commit
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    def test_10(self):
        with db_session:
            g2 = Group.get(number=102)
            s1 = Subject.get(name='Subj1')
            g2.subjects.add(s1)
            g2.subjects.clear()
            last_sql = db.last_sql

        with db_session:
            self.assertEqual(db.last_sql, last_sql)  # assert no DELETE statement on commit
            db_subjects = db.select('subject from Group_Subject where "group" = 102')
            self.assertEqual(db_subjects , [])

    def test_11(self):
        with db_session:
            g = Group.get(number=101)
            s1, s2, s3, s4 = Subject.select()[:]
            g.subjects = [ s2, s3 ]

        with db_session:
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj2', 'Subj3'})

    def test_12(self):
        with db_session:
            g = Group.get(number=101)
            s1, s2, s3, s4 = Subject.select()[:]
            g.subjects.remove(s2)
            g.subjects = [ s1, s2 ]
            last_sql = db.last_sql

        with db_session:
            self.assertEqual(db.last_sql, last_sql)  # assert no INSERT statement on commit
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    def test_13(self):
        with db_session:
            g = Group.get(number=101)
            s1, s2, s3, s4 = Subject.select()[:]
            g.subjects.add(s3)
            g.subjects = [ s1, s2 ]
            last_sql = db.last_sql

        with db_session:
            self.assertEqual(db.last_sql, last_sql)  # assert no DELETE statement on commit
            db_subjects = db.select('subject from Group_Subject where "group" = 101')
            self.assertEqual(set(db_subjects), {'Subj1', 'Subj2'})

    @db_session
    def test_14(self):
        g1 = Group[101]
        s1 = Subject['Subj1']
        self.assertTrue(s1 in g1.subjects)

        group_setdata = g1._vals_[Group.subjects]
        self.assertTrue(s1 in group_setdata)
        self.assertEqual(group_setdata.added, None)
        self.assertEqual(group_setdata.removed, None)

        subj_setdata = s1._vals_[Subject.groups]
        self.assertTrue(g1 in subj_setdata)
        self.assertEqual(subj_setdata.added, None)
        self.assertEqual(subj_setdata.removed, None)

        g1.subjects.remove(s1)
        self.assertTrue(s1 not in group_setdata)
        self.assertEqual(group_setdata.added, None)
        self.assertEqual(group_setdata.removed, {s1})
        self.assertTrue(g1 not in subj_setdata)
        self.assertEqual(subj_setdata.added, None)
        self.assertEqual(subj_setdata.removed, {g1})

        g1.subjects.add(s1)
        self.assertTrue(s1 in group_setdata)
        self.assertEqual(group_setdata.added, set())
        self.assertEqual(group_setdata.removed, set())
        self.assertTrue(g1 in subj_setdata)
        self.assertEqual(subj_setdata.added, set())
        self.assertEqual(subj_setdata.removed, set())

    @db_session
    def test_15(self):
        g = Group[101]
        e = g.subjects.is_empty()
        self.assertEqual(e, False)

        db._dblocal.last_sql = None
        e = g.subjects.is_empty()  # should take result from the cache
        self.assertEqual(e, False)
        self.assertEqual(db.last_sql, None)

        g = Group[102]
        e = g.subjects.is_empty()  # should take SQL from the SQL cache
        self.assertEqual(e, True)

        db._dblocal.last_sql = None
        e = g.subjects.is_empty()  # should take result from the cache
        self.assertEqual(e, True)
        self.assertEqual(db.last_sql, None)

    @db_session
    def test_16(self):
        g = Group[101]
        c = len(g.subjects)
        self.assertEqual(c, 2)
        db._dblocal.last_sql = None
        e = g.subjects.is_empty()  # should take result from the cache
        self.assertEqual(e, False)
        self.assertEqual(db.last_sql, None)

        g = Group[102]
        c = len(g.subjects)
        self.assertEqual(c, 0)
        db._dblocal.last_sql = None
        e = g.subjects.is_empty()  # should take result from the cache
        self.assertEqual(e, True)
        self.assertEqual(db.last_sql, None)

    @db_session
    def test_17(self):
        g = Group[101]
        s1 = Subject['Subj1']
        s3 = Subject['Subj3']
        c = g.subjects.count()
        self.assertEqual(c, 2)

        db._dblocal.last_sql = None
        c = g.subjects.count()  # should take count from the cache
        self.assertEqual(c, 2)
        self.assertEqual(db.last_sql, None)

        g.subjects.add(s3)
        db._dblocal.last_sql = None
        c = g.subjects.count()  # should take modified count from the cache
        self.assertEqual(c, 3)
        self.assertEqual(db.last_sql, None)

        g.subjects.remove(s1)
        db._dblocal.last_sql = None
        c = g.subjects.count()  # should take modified count from the cache
        self.assertEqual(c, 2)
        self.assertEqual(db.last_sql, None)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, print_function, division

import unittest

from foalorm.orm.core import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

db = Database()


class Person(db.Entity):
    name = Required(str)


class TestFlush(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)

    @classmethod
    def tearDownClass(self):
        teardown_database(db)

    def test1(self):
        with db_session:
            a = Person(name='A')
            b = Person(name='B')
            c = Person(name='C')
            self.assertEqual(a.id, None)
            self.assertEqual(b.id, None)
            self.assertEqual(c.id, None)

            b.flush()
            self.assertEqual(a.id, None)
            self.assertIsNotNone(b.id)
            b_id = b.id
            self.assertEqual(c.id, None)

            flush()
            self.assertIsNotNone(a.id)
            self.assertEqual(b.id, b_id)
            self.assertIsNotNone(c.id)
            self.assertEqual(len({a.id, b.id, c.id}), 3)



import unittest

from foalorm.orm import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database


class TestCascade(unittest.TestCase):
    providers = ['sqlite']  # Implement for other providers

    def tearDown(self):
        if self.db.schema is not None:
            teardown_database(self.db)

    def assert_on_delete(self, table_name, value):
        db = self.db
        if not (db.provider.dialect == 'SQLite' and foalorm.__version__ < '0.9'):
            table_name = table_name.lower()
        table = db.schema.tables[table_name]
        fkeys = table.foreign_keys
        self.assertEqual(1, len(fkeys))
        if foalorm.__version__ >= '0.9':
            self.assertEqual(fkeys[0].on_delete, value)
        elif db.provider.dialect == 'SQLite':
            self.assertIn('ON DELETE %s' % value, table.get_create_command())
        else:
            self.assertIn('ON DELETE %s' % value, list(fkeys.values())[0].get_create_command())


    def test_1(self):
        db = self.db = Database()

        class Person(self.db.Entity):
            name = Required(str)
            group = Required('Group')

        class Group(self.db.Entity):
            persons = Set(Person)

        setup_database(db)
        self.assert_on_delete('Person', 'CASCADE')

    def test_2(self):
        db = self.db = Database()

        class Person(self.db.Entity):
            name = Required(str)
            group = Required('Group')

        class Group(self.db.Entity):
            persons = Set(Person, cascade_delete=True)

        setup_database(db)
        self.assert_on_delete('Person', 'CASCADE')

    def test_3(self):
        db = self.db = Database()

        class Person(self.db.Entity):
            name = Required(str)
            group = Optional('Group')

        class Group(self.db.Entity):
            persons = Set(Person, cascade_delete=True)

        setup_database(db)
        self.assert_on_delete('Person', 'CASCADE')

    @raises_exception(TypeError, "'cascade_delete' option cannot be set for attribute Group.persons, because reverse attribute Person.group is collection")
    def test_4(self):
        db = self.db = Database()

        class Person(self.db.Entity):
            name = Required(str)
            group = Set('Group')

        class Group(self.db.Entity):
            persons = Set(Person, cascade_delete=True)

        setup_database(db)

    @raises_exception(TypeError, "'cascade_delete' option cannot be set for both sides of relationship (Person.group and Group.persons) simultaneously")
    def test_5(self):
        db = self.db = Database()

        class Person(self.db.Entity):
            name = Required(str)
            group = Set('Group', cascade_delete=True)

        class Group(self.db.Entity):
            persons = Required(Person, cascade_delete=True)

        setup_database(db)

    def test_6(self):
        db = self.db = Database()

        class Person(self.db.Entity):
            name = Required(str)
            group = Set('Group')

        class Group(self.db.Entity):
            persons = Optional(Person)

        setup_database(db)
        self.assert_on_delete('Group', 'SET NULL')


if __name__ == '__main__':
    unittest.main()



import unittest

from foalorm.orm import *
from foalorm.orm.tests import setup_database, teardown_database

class EntityStatusTestCase(object):
    @classmethod
    def setUpClass(cls):
        db = cls.db = Database()

        class TestPost(db.Entity):
            category = Optional('TestCategory')
            name = Optional(str, default='Noname')

        class TestCategory(db.Entity):
            posts = Set(TestPost)

        setup_database(db)

    @classmethod
    def tearDownClass(cls):
        teardown_database(cls.db)

    def make_flush(self, obj=None):
        raise NotImplementedError

    @db_session
    def test_delete_updated(self):
        p = self.db.TestPost()
        self.make_flush(p)
        p.name = 'FoalORM'
        self.assertEqual(p._status_, 'modified')
        self.make_flush(p)
        self.assertEqual(p._status_, 'updated')
        p.delete()
        self.assertEqual(p._status_, 'marked_to_delete')
        self.make_flush(p)
        self.assertEqual(p._status_, 'deleted')

    @db_session
    def test_delete_inserted(self):
        p = self.db.TestPost()
        self.assertEqual(p._status_, 'created')
        self.make_flush(p)
        self.assertEqual(p._status_, 'inserted')
        p.delete()

    @db_session
    def test_cancelled(self):
        p = self.db.TestPost()
        self.assertEqual(p._status_, 'created')
        p.delete()
        self.assertEqual(p._status_, 'cancelled')
        self.make_flush(p)
        self.assertEqual(p._status_, 'cancelled')


class EntityStatusTestCase_ObjectFlush(EntityStatusTestCase,
                                       unittest.TestCase):

    def make_flush(self, obj=None):
        obj.flush()


class EntityStatusTestCase_FullFlush(EntityStatusTestCase,
                                     unittest.TestCase):

    def make_flush(self, obj=None):
        flush() # full flush



from __future__ import absolute_import, print_function, division

import unittest

from foalorm import orm
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database

class TestIntConverter1(unittest.TestCase):
    def setUp(self):
        self.db = db = orm.Database()

        class Foo(db.Entity):
            id = orm.PrimaryKey(int)
            x = orm.Required(int, size=8, unsigned=True)

        setup_database(db)

        with orm.db_session:
            foo = Foo(id=123, x=1)

    def tearDown(self):
        teardown_database(self.db)

    def test_1(self):
        with orm.db_session:
            foo = self.db.Foo[123]
            foo.x -= 1
        with orm.db_session:
            foo = self.db.Foo[123]
            self.assertEqual(foo.x, 0)

    @raises_exception(ValueError, "Value -1 of attr Foo.x is less than the minimum allowed value 0")
    @orm.db_session
    def test_2(self):
        foo = self.db.Foo[123]
        foo.x -= 2

    @orm.db_session
    def test_3(self):
        with orm.db_session:
            foo = self.db.Foo[123]
            foo.x += 254
        with orm.db_session:
            foo = self.db.Foo[123]
            self.assertEqual(foo.x, 255)

    @raises_exception(ValueError, "Value 256 of attr Foo.x is greater than the maximum allowed value 255")
    @orm.db_session
    def test_4(self):
        foo = self.db.Foo[123]
        foo.x += 255



import unittest

from foalorm.orm.ormtypes import TrackedList, TrackedDict, TrackedValue

class Object(object):
    def __init__(self):
        self.on_attr_changed = None
    def _attr_changed_(self, attr):
        if self.on_attr_changed is not None:
            self.on_attr_changed(attr)


class Attr(object):
    pass


class TestTrackedValue(unittest.TestCase):

    def test_make(self):
        obj = Object()
        attr = Attr()
        value = {'items': ['one', 'two', 'three']}
        tracked_value = TrackedValue.make(obj, attr, value)
        self.assertEqual(type(tracked_value), TrackedDict)
        self.assertEqual(type(tracked_value['items']), TrackedList)

    def test_dict_setitem(self):
        obj = Object()
        attr = Attr()
        value = {'items': ['one', 'two', 'three']}
        tracked_value = TrackedValue.make(obj, attr, value)
        log = []
        obj.on_attr_changed = lambda x: log.append(x)
        tracked_value['items'] = [1, 2, 3]
        self.assertEqual(log, [attr])

    def test_list_append(self):
        obj = Object()
        attr = Attr()
        value = {'items': ['one', 'two', 'three']}
        tracked_value = TrackedValue.make(obj, attr, value)
        log = []
        obj.on_attr_changed = lambda x: log.append(x)
        tracked_value['items'].append('four')
        self.assertEqual(log, [attr])

    def test_list_setslice(self):
        obj = Object()
        attr = Attr()
        value = {'items': ['one', 'two', 'three']}
        tracked_value = TrackedValue.make(obj, attr, value)
        log = []
        obj.on_attr_changed = lambda x: log.append(x)
        tracked_value['items'][1:2] = ['a', 'b', 'c']
        self.assertEqual(log, [attr])
        self.assertEqual(tracked_value['items'], ['one', 'a', 'b', 'c', 'three'])



import sys, platform

PYPY = platform.python_implementation() == 'PyPy'
PY37 = sys.version_info[:2] >= (3, 7)
PY38 = sys.version_info[:2] >= (3, 8)
PY39 = sys.version_info[:2] >= (3, 9)
PY310 = sys.version_info[:2] >= (3, 10)
PY311 = sys.version_info[:2] >= (3, 11)

unicode = str
buffer = bytes
int_types = (int,)

def cmp(a, b):
    return (a > b) - (a < b)



from __future__ import absolute_import, print_function, division

import unittest
from datetime import datetime, date, time

from foalorm.orm.tests.testutils import raises_exception
from foalorm.converting import str2time


class TestTimeParsing(unittest.TestCase):
    def test_time_1(self):
        self.assertEqual(str2time('1:2'), time(1, 2))
        self.assertEqual(str2time('01:02'), time(1, 2))
        self.assertEqual(str2time('1:2:3'), time(1, 2, 3))
        self.assertEqual(str2time('01:02:03'), time(1, 2, 3))
        self.assertEqual(str2time('1:2:3.4'), time(1, 2, 3, 400000))
        self.assertEqual(str2time('01:02:03.4'), time(1, 2, 3, 400000))

    @raises_exception(ValueError, 'Unrecognized time format')
    def test_time_2(self):
        str2time('1:')

    @raises_exception(ValueError, 'Unrecognized time format')
    def test_time_3(self):
        str2time('1: 2')

    @raises_exception(ValueError, 'Unrecognized time format')
    def test_time_4(self):
        str2time('1:2:')

    @raises_exception(ValueError, 'Unrecognized time format')
    def test_time_5(self):
        str2time('1:2:3:')

    @raises_exception(ValueError, 'Unrecognized time format')
    def test_time_6(self):
        str2time('1:2:3.1234567')

    def test_time_7(self):
        self.assertEqual(str2time('1:33 am'), time(1, 33))
        self.assertEqual(str2time('2:33 am'), time(2, 33))
        self.assertEqual(str2time('11:33 am'), time(11, 33))
        self.assertEqual(str2time('12:33 am'), time(0, 33))

    def test_time_8(self):
        self.assertEqual(str2time('1:33 pm'), time(13, 33))
        self.assertEqual(str2time('2:33 pm'), time(14, 33))
        self.assertEqual(str2time('11:33 pm'), time(23, 33))
        self.assertEqual(str2time('12:33 pm'), time(12, 33))

    def test_time_9(self):
        self.assertEqual(str2time('1:33am'), time(1, 33))
        self.assertEqual(str2time('1:33 am'), time(1, 33))
        self.assertEqual(str2time('1:33 AM'), time(1, 33))
        self.assertEqual(str2time('1:33 a.m'), time(1, 33))
        self.assertEqual(str2time('1:33 A.M'), time(1, 33))
        self.assertEqual(str2time('1:33 a.m.'), time(1, 33))
        self.assertEqual(str2time('1:33 A.M.'), time(1, 33))

    def test_time_10(self):
        self.assertEqual(str2time('1:33pm'), time(13, 33))
        self.assertEqual(str2time('1:33 pm'), time(13, 33))
        self.assertEqual(str2time('1:33 PM'), time(13, 33))
        self.assertEqual(str2time('1:33 p.m'), time(13, 33))
        self.assertEqual(str2time('1:33 P.M'), time(13, 33))
        self.assertEqual(str2time('1:33 p.m.'), time(13, 33))
        self.assertEqual(str2time('1:33 P.M.'), time(13, 33))

    def test_time_11(self):
        self.assertEqual(str2time('12:34:56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12.34.56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12 34 56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12h34m56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12h 34m 56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12 h 34 m 56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12h 34m 56.789s'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12 h 34 m 56.789 s'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12h 34min 56.789'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12h 34min 56.789sec'), time(12, 34, 56, 789000))
        self.assertEqual(str2time('12h 34 min 56.789 sec'), time(12, 34, 56, 789000))

    def test_time_12(self):
        self.assertEqual(str2time('12:34:56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12.34.56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12 34 56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12h34m56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12h 34m 56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12 h 34 m 56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12h 34m 56.789s a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12 h 34 m 56.789 s a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12h 34min 56.789 a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12h 34min 56.789sec a.m.'), time(0, 34, 56, 789000))
        self.assertEqual(str2time('12h 34 min 56.789 sec a.m.'), time(0, 34, 56, 789000))

    def test_time_13(self):
        self.assertEqual(str2time('12:34'), time(12, 34))
        self.assertEqual(str2time('12.34'), time(12, 34))
        self.assertEqual(str2time('12 34'), time(12, 34))
        self.assertEqual(str2time('12h34'), time(12, 34))
        self.assertEqual(str2time('12h34m'), time(12, 34))
        self.assertEqual(str2time('12h 34m'), time(12, 34))
        self.assertEqual(str2time('12h34min'), time(12, 34))
        self.assertEqual(str2time('12h 34min'), time(12, 34))
        self.assertEqual(str2time('12 h 34 m'), time(12, 34))
        self.assertEqual(str2time('12 h 34 min'), time(12, 34))
        self.assertEqual(str2time('12u34'), time(12, 34))  # Belgium
        self.assertEqual(str2time('12h'), time(12))
        self.assertEqual(str2time('12u'), time(12))

    def test_time_14(self):
        self.assertEqual(str2time('12:34 am'), time(0, 34))
        self.assertEqual(str2time('12.34 am'), time(0, 34))
        self.assertEqual(str2time('12 34 am'), time(0, 34))
        self.assertEqual(str2time('12h34 am'), time(0, 34))
        self.assertEqual(str2time('12h34m am'), time(0, 34))
        self.assertEqual(str2time('12h 34m am'), time(0, 34))
        self.assertEqual(str2time('12h34min am'), time(0, 34))
        self.assertEqual(str2time('12h 34min am'), time(0, 34))
        self.assertEqual(str2time('12 h 34 m am'), time(0, 34))
        self.assertEqual(str2time('12 h 34 min am'), time(0, 34))
        self.assertEqual(str2time('12u34 am'), time(0, 34))
        self.assertEqual(str2time('12h am'), time(0))
        self.assertEqual(str2time('12u am'), time(0))


if __name__ == '__main__':
    unittest.main()



import unittest
from decimal import Decimal
from datetime import datetime, time
from random import randint

from foalorm import orm
from foalorm.orm.core import *
from foalorm.orm.tests import setup_database, teardown_database
from foalorm.orm.tests.testutils import raises_exception

db = Database()

class Person(db.Entity):
    id = PrimaryKey(int)
    name = orm.Required(str, 40)
    lastName = orm.Required(str, max_len=40, unique=True)
    age = orm.Optional(int, max=60, min=10)
    nickName = orm.Optional(str, autostrip=False)
    middleName = orm.Optional(str, nullable=True)
    rate = orm.Optional(Decimal, precision=11)
    salaryRate = orm.Optional(Decimal, precision=13, scale=8)
    timeStmp = orm.Optional(datetime, precision=6)
    gpa = orm.Optional(float, py_check=lambda val: val >= 0 and val <= 5)
    vehicle = orm.Optional(str, column='car')

class TestAttributeOptions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with orm.db_session:
            p1 = Person(id=1, name='Andrew', lastName='Bodroue', age=40, rate=0.980000000001, salaryRate=0.98000001)
            p2 = Person(id=2, name='Vladimir', lastName='Andrew ', nickName='vlad  ')
            p3 = Person(id=3, name='Nick', lastName='Craig', middleName=None, timeStmp='2010-12-10 14:12:09.019473',
                        vehicle='dodge')

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()

    def tearDown(self):
        rollback()
        db_session.__exit__()
    
    def test_optionalStringEmpty(self):
        queryResult = select(p.id for p in Person if p.nickName==None).first()
        self.assertIsNone(queryResult)
        
    def test_optionalStringNone(self):
        queryResult = select(p.id for p in Person if p.middleName==None).first()
        self.assertIsNotNone(queryResult)        
    
    def test_stringAutoStrip(self):
        self.assertEqual(Person[2].lastName, 'Andrew')
        
    def test_stringAutoStripFalse(self):
        self.assertEqual(Person[2].nickName, 'vlad  ')
    
    def test_intNone(self):
        queryResult = select(p.id for p in Person if p.age==None).first()
        self.assertIsNotNone(queryResult)
    
    def test_columnName(self):
        self.assertEqual(getattr(Person.vehicle, 'column'), 'car')
        
    def test_decimalPrecisionTwo(self):
        queryResult = select(p.rate for p in Person if p.age==40).first()
        self.assertAlmostEqual(float(queryResult), 0.98, 12)
    
    def test_decimalPrecisionEight(self):
        queryResult = select(p.salaryRate for p in Person if p.age==40).first()
        self.assertAlmostEqual(float(queryResult), 0.98000001, 8)
        
    def test_fractionalSeconds(self):
        queryResult = select(p.timeStmp for p in Person if p.name=='Nick').first()
        self.assertEqual(queryResult.microsecond, 19473)
    
    def test_intMax(self):
        p4 = Person(id=4, name='Denis', lastName='Blanc', age=60)
        
    def test_intMin(self):
        p4 = Person(id=4, name='Denis', lastName='Blanc', age=10)
    
    @raises_exception(ValueError, "Value 61 of attr Person.age is greater than the maximum allowed value 60")
    def test_intMaxException(self):
        p4 = Person(id=4, name='Denis', lastName='Blanc', age=61)
     
    @raises_exception(ValueError, "Value 9 of attr Person.age is less than the minimum allowed value 10")   
    def test_intMinException(self):
        p4 = Person(id=4, name='Denis', lastName='Blanc', age=9)

    def test_py_check(self):
        p4 = Person(id=4, name='Denis', lastName='Blanc', gpa=5)
        p5 = Person(id=5, name='Mario', lastName='Gon', gpa=1)
        flush()

    @raises_exception(ValueError, "Check for attribute Person.gpa failed. Value: 6.0")
    def test_py_checkMoreException(self):        
        p6 = Person(id=6, name='Daniel', lastName='Craig', gpa=6)

    @raises_exception(ValueError, "Check for attribute Person.gpa failed. Value: -1.0")
    def test_py_checkLessException(self):        
        p6 = Person(id=6, name='Daniel', lastName='Craig', gpa=-1)
    
    @raises_exception(TransactionIntegrityError,
                      'Object Person[...] cannot be stored in the database. IntegrityError: ...')
    def test_unique(self):
        p6 = Person(id=6, name='Boris', lastName='Bodroue')
        flush()



from __future__ import absolute_import, print_function, division
from foalorm.py23compat import buffer, int_types

import sys, types, weakref
from decimal import Decimal
from datetime import date, time, datetime, timedelta
from functools import wraps
from uuid import UUID

from foalorm.utils import throw, parse_expr, deref_proxy

NoneType = type(None)

class LongStr(str):
    lazy = True

LongUnicode = LongStr

class SetType(object):
    __slots__ = 'item_type'
    def __deepcopy__(self, memo):
        return self  # SetType instances are "immutable"
    def __init__(self, item_type):
        self.item_type = item_type
    def __eq__(self, other):
        return type(other) is SetType and self.item_type == other.item_type
    def __ne__(self, other):
        return type(other) is not SetType or self.item_type != other.item_type
    def __hash__(self):
        return hash(self.item_type) + 1

class FuncType(object):
    __slots__ = 'func'
    def __deepcopy__(self, memo):
        return self  # FuncType instances are "immutable"
    def __init__(self, func):
        self.func = func
    def __eq__(self, other):
        return type(other) is FuncType and self.func == other.func
    def __ne__(self, other):
        return type(other) is not FuncType or self.func != other.func
    def __hash__(self):
        return hash(self.func) + 1
    def __repr__(self):
        return 'FuncType(%s at %d)' % (self.func.__name__, id(self.func))

class MethodType(object):
    __slots__ = 'obj', 'func'
    def __deepcopy__(self, memo):
        return self  # MethodType instances are "immutable"
    def __init__(self, method):
        self.obj = method.__self__
        self.func = method.__func__
    def __eq__(self, other):
        return type(other) is MethodType and self.obj == other.obj and self.func == other.func
    def __ne__(self, other):
        return type(other) is not MethodType or self.obj != other.obj or self.func != other.func
    def __hash__(self):
        return hash(self.obj) ^ hash(self.func)

raw_sql_cache = {}

def parse_raw_sql(sql):
    result = raw_sql_cache.get(sql)
    if result is not None: return result
    if not isinstance(sql, str) or not sql:
        throw(TypeError, "Raw SQL string fragment expected. Got: %r" % sql)
    items = []
    codes = []
    pos = 0
    while True:
        try: i = sql.index('$', pos)
        except ValueError:
            items.append(sql[pos:])
            break
        items.append(sql[pos:i])
        if sql[i+1] == '$':
            items.append('$')
            pos = i+2
        else:
            try: expr, _ = parse_expr(sql, i+1)
            except ValueError:
                raise ValueError(sql[i:])
            pos = i+1 + len(expr)
            if expr.endswith(';'): expr = expr[:-1]
            code = compile(expr, '<?>', 'eval')  # expr correction check
            codes.append(code)
            items.append((expr, code))
    result = tuple(items), tuple(codes)
    raw_sql_cache[sql] = result
    return result

def raw_sql(sql, result_type=None):
    globals = sys._getframe(1).f_globals
    locals = sys._getframe(1).f_locals
    return RawSQL(sql, globals, locals, result_type)

class RawSQL(object):
    def __deepcopy__(self, memo):
        assert False  # should not attempt to deepcopy RawSQL instances, because of locals/globals
    def __init__(self, sql, globals=None, locals=None, result_type=None):
        self.sql = sql
        self.items, self.codes = parse_raw_sql(sql)
        self.types, self.values = normalize(tuple(eval(code, globals, locals) for code in self.codes))
        self.result_type = result_type
    def _get_type_(self):
        return RawSQLType(self.sql, self.items, self.types, self.result_type)

class RawSQLType(object):
    def __deepcopy__(self, memo):
        return self  # RawSQLType instances are "immutable"
    def __init__(self, sql, items, types, result_type):
        self.sql = sql
        self.items = items
        self.types = types
        self.result_type = result_type
    def __hash__(self):
        return hash(self.sql) ^ hash(self.types)
    def __eq__(self, other):
        return type(other) is RawSQLType and self.sql == other.sql and self.types == other.types
    def __ne__(self, other):
        return not self.__eq__(other)

class QueryType(object):
    def __init__(self, query, limit=None, offset=None):
        self.query_key = query._key
        self.translator = query._translator
        self.limit = limit
        self.offset = offset
    def __hash__(self):
        result = hash(self.query_key)
        if self.limit is not None:
            result ^= hash(self.limit + 3)
        if self.offset is not None:
            result ^= hash(self.offset)
        return result
    def __eq__(self, other):
        return type(other) is QueryType and self.query_key == other.query_key \
               and self.limit == other.limit and self.offset == other.offset
    def __ne__(self, other):
        return not self.__eq__(other)


def normalize(value):
    value = deref_proxy(value)
    t = type(value)
    if t is tuple:
        item_types, item_values = [], []
        for item in value:
            item_type, item_value = normalize(item)
            item_values.append(item_value)
            item_types.append(item_type)
        return tuple(item_types), tuple(item_values)

    if t.__name__ == 'EntityMeta':
        return SetType(value), value

    if t.__name__ == 'EntityIter':
        entity = value.entity
        return SetType(entity), entity

    if isinstance(value, str):
        return str, value

    if t in function_types:
        return FuncType(value), value

    if t is types.MethodType:
        return MethodType(value), value

    if hasattr(value, '_get_type_'):
        return value._get_type_(), value

    return normalize_type(t), value

def normalize_type(t):
    tt = type(t)
    if tt is tuple: return tuple(normalize_type(item) for item in t)
    if not isinstance(t, type):
        return t
    assert t.__name__ != 'EntityMeta'
    if tt.__name__ == 'EntityMeta': return t
    if t is NoneType: return t
    t = type_normalization_dict.get(t, t)
    if t in primitive_types: return t
    if t in (slice, type(Ellipsis)): return t
    if issubclass(t, str): return str
    if issubclass(t, (dict, Json)): return Json
    if issubclass(t, Array): return t
    throw(TypeError, 'Unsupported type %r' % t.__name__)

coercions = {
    (int, float): float,
    (int, Decimal): Decimal,
    (date, datetime): datetime,
    (bool, int): int,
    (bool, float): float,
    (bool, Decimal): Decimal
    }
coercions.update(((t2, t1), t3) for ((t1, t2), t3) in list(coercions.items()))

def coerce_types(t1, t2):
    if t1 == t2: return t1
    is_set_type = False
    if type(t1) is SetType:
        is_set_type = True
        t1 = t1.item_type
    if type(t2) is SetType:
        is_set_type = True
        t2 = t2.item_type
    result = coercions.get((t1, t2))
    if result is not None and is_set_type: result = SetType(result)
    return result

def are_comparable_types(t1, t2, op='=='):
    # types must be normalized already!
    tt1 = type(t1)
    tt2 = type(t2)

    t12 = {t1, t2}
    if Json in t12 and t12 < {Json, str, str, int, bool, float}:
        return True
    if op in ('in', 'not in'):
        if tt2 is RawSQLType: return True
        if tt2 is not SetType: return False
        op = '=='
        t2 = t2.item_type
        tt2 = type(t2)
    if op in ('is', 'is not'):
        return t1 is not None and t2 is NoneType
    if tt1 is tuple:
        if not tt2 is tuple: return False
        if len(t1) != len(t2): return False
        for item1, item2 in zip(t1, t2):
            if not are_comparable_types(item1, item2): return False
        return True
    if tt1 is RawSQLType or tt2 is RawSQLType: return True
    if op in ('==', '<>', '!='):
        if t1 is NoneType and t2 is NoneType: return False
        if t1 is NoneType or t2 is NoneType: return True
        if t1 in primitive_types:
            if t1 is t2: return True
            if (t1, t2) in coercions: return True
            if tt1 is not type or tt2 is not type: return False
            if issubclass(t1, int_types) and issubclass(t2, str): return True
            if issubclass(t2, int_types) and issubclass(t1, str): return True
            return False
        if tt1.__name__ == tt2.__name__ == 'EntityMeta':
            return t1._root_ is t2._root_
        return False
    if t1 is t2 and t1 in comparable_types: return True
    return (t1, t2) in coercions

class TrackedValue(object):
    def __init__(self, obj, attr):
        self.obj_ref = weakref.ref(obj)
        self.attr = attr
    @classmethod
    def make(cls, obj, attr, value):
        if isinstance(value, dict):
            return TrackedDict(obj, attr, value)
        if isinstance(value, list):
            return TrackedList(obj, attr, value)
        return value
    def _changed_(self):
        obj = self.obj_ref()
        if obj is not None:
            obj._attr_changed_(self.attr)
    def get_untracked(self):
        assert False, 'Abstract method'  # pragma: no cover

def tracked_method(func):
    @wraps(func)
    def new_func(self, *args, **kwargs):
        obj = self.obj_ref()
        attr = self.attr
        if obj is not None:
            args = tuple(TrackedValue.make(obj, attr, arg) for arg in args)
            if kwargs: kwargs = {key: TrackedValue.make(obj, attr, value) for key, value in kwargs.items()}
        result = func(self, *args, **kwargs)
        self._changed_()
        return result
    return new_func

class TrackedDict(TrackedValue, dict):
    def __init__(self, obj, attr, value):
        TrackedValue.__init__(self, obj, attr)
        dict.__init__(self, {key: self.make(obj, attr, val) for key, val in value.items()})
    def __reduce__(self):
        return dict, (dict(self),)
    __setitem__ = tracked_method(dict.__setitem__)
    __delitem__ = tracked_method(dict.__delitem__)
    _update = tracked_method(dict.update)
    def update(self, *args, **kwargs):
        args = [ arg if isinstance(arg, dict) else dict(arg) for arg in args ]
        return self._update(*args, **kwargs)
    setdefault = tracked_method(dict.setdefault)
    pop = tracked_method(dict.pop)
    popitem = tracked_method(dict.popitem)
    clear = tracked_method(dict.clear)
    def get_untracked(self):
        return {key: val.get_untracked() if isinstance(val, TrackedValue) else val
                for key, val in self.items()}

class TrackedList(TrackedValue, list):
    def __init__(self, obj, attr, value):
        TrackedValue.__init__(self, obj, attr)
        list.__init__(self, (self.make(obj, attr, val) for val in value))
    def __reduce__(self):
        return list, (list(self),)
    __setitem__ = tracked_method(list.__setitem__)
    __delitem__ = tracked_method(list.__delitem__)
    extend = tracked_method(list.extend)
    append = tracked_method(list.append)
    pop = tracked_method(list.pop)
    remove = tracked_method(list.remove)
    insert = tracked_method(list.insert)
    reverse = tracked_method(list.reverse)
    sort = tracked_method(list.sort)
    clear = tracked_method(list.clear)
    def get_untracked(self):
        return [val.get_untracked() if isinstance(val, TrackedValue) else val for val in self]

def validate_item(item_type, item):
    if not isinstance(item, item_type):
        if item_type is not str and hasattr(item, '__index__'):
            return item.__index__()
        throw(TypeError, 'Cannot store %r item in array of %r' % (type(item).__name__, item_type.__name__))
    return item

class TrackedArray(TrackedList):
    def __init__(self, obj, attr, value):
        TrackedList.__init__(self, obj, attr, value)
        self.item_type = attr.py_type.item_type
    def extend(self, items):
        items = [validate_item(self.item_type, item) for item in items]
        TrackedList.extend(self, items)
    def append(self, item):
        item = validate_item(self.item_type, item)
        TrackedList.append(self, item)
    def insert(self, index, item):
        item = validate_item(self.item_type, item)
        TrackedList.insert(self, index, item)
    def __setitem__(self, index, item):
        item = validate_item(self.item_type, item)
        TrackedList.__setitem__(self, index, item)

    def __contains__(self, item):
        if not isinstance(item, str) and hasattr(item, '__iter__'):
            return all(it in set(self) for it in item)
        return list.__contains__(self, item)


class Json(object):
    """A wrapper over a dict or list
    """
    @classmethod
    def default_empty_value(cls):
        return {}

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __repr__(self):
        return '<Json %r>' % self.wrapped

class Array(object):
    item_type = None  # Should be overridden in subclass

    @classmethod
    def default_empty_value(cls):
        return []


class IntArray(Array):
    item_type = int


class StrArray(Array):
    item_type = str


class FloatArray(Array):
    item_type = float


numeric_types = {bool, int, float, Decimal}
comparable_types = {int, float, Decimal, str, date, time, datetime, timedelta, bool, UUID, IntArray, StrArray, FloatArray}
primitive_types = comparable_types | {buffer}
function_types = {type, types.FunctionType, types.BuiltinFunctionType}
type_normalization_dict = {}

array_types = {
    int: IntArray,
    float: FloatArray,
    str: StrArray
}




import unittest
from foalorm.orm.core import *
from foalorm.orm.tests.testutils import *
from foalorm.orm.tests import setup_database, teardown_database


db = Database()

class Person(db.Entity):
    first_name = Required(str)
    last_name = Required(str)
    age = Optional(int)
    value = Required(float)


class TestFString(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        setup_database(db)
        with db_session:
            Person(id=1, first_name='Alexander', last_name='Tischenko', age=23, value=1.4)
            Person(id=2, first_name='Alexander', last_name='Kozlovskiy', age=42, value=1.2)
            Person(id=3, first_name='Arthur', last_name='Pendragon', age=54, value=1.33)
            Person(id=4, first_name='Okita', last_name='Souji', age=15, value=2.1)
            Person(id=5, first_name='Musashi', last_name='Miyamoto', age=None, value=0.9)
            Person(id=6, first_name='Jeanne', last_name="d'Arc", age=30, value=43.212)

    @classmethod
    def tearDownClass(cls):
        teardown_database(db)

    def setUp(self):
        rollback()
        db_session.__enter__()
    def tearDown(self):
        rollback()
        db_session.__exit__()

    def test_1(self):
        x = 'Alexander'
        y = 'Tischenko'

        q = select(p.id for p in Person if p.first_name + ' ' + p.last_name == f'{x} {y}')
        self.assertEqual(set(q), {1})

        q2 = select("""p.id for p in Person if p.first_name + ' ' + p.last_name == f'{x} {y}'""")
        self.assertEqual(set(q2), {1})

    def test_2(self):
        q = select(p.id for p in Person if f'{p.first_name} {p.last_name}' == 'Alexander Tischenko')
        self.assertEqual(set(q), {1})

        q2 = select("""p.id for p in Person if f'{p.first_name} {p.last_name}' == 'Alexander Tischenko'""")
        self.assertEqual(set(q2), {1})

    def test_3(self):
        x = 'Great'

        q = select(f'{p.first_name!s} the {x}' for p in Person if p.id == 1)
        self.assertEqual(set(q), {'Alexander the Great'})

        q2 = select("""f'{p.first_name!s} the {x}' for p in Person if p.id == 1""")
        self.assertEqual(set(q2), {'Alexander the Great'})

    def test_4(self):
        q = select(f'{p.first_name} {p.age}' for p in Person if p.id == 1)
        self.assertEqual(set(q), {'Alexander 23'})

        q2 = select("""f'{p.first_name} {p.age}' for p in Person if p.id == 1""")
        self.assertEqual(set(q2), {'Alexander 23'})

    def test_5(self):
        q = select(f'{p.first_name} {p.age}' for p in Person if p.id == 1)
        self.assertEqual(set(q), {'Alexander 23'})

        q2 = select("""f'{p.first_name} {p.age}' for p in Person if p.id == 1""")
        self.assertEqual(set(q2), {'Alexander 23'})

    @raises_exception(NotImplementedError, 'You cannot set width and precision for f-string expression in query')
    def test_6a(self):
        width = 3
        precision = 4
        q = select(p.id for p in Person if f'{p.value:{width}.{precision}}')[:]

    @raises_exception(NotImplementedError, 'You cannot set width and precision for f-string expression in query')
    def test_6b(self):
        width = 3
        precision = 4
        q = select("""p.id for p in Person if f'{p.value:{width}.{precision}}'""")[:]
        q.show()

    @raises_exception(NotImplementedError, 'You cannot specify conversion type for f-string expression in query')
    def test_7a(self):
        width = 3
        precision = 4
        q = select(p.id for p in Person if f'{p.value!r}')[:]

    @raises_exception(NotImplementedError, 'You cannot specify conversion type for f-string expression in query')
    def test_7b(self):
        width = 3
        precision = 4
        q = select("""p.id for p in Person if f'{p.value!r}'""")[:]
        q.show()

    def test_8(self):
        x = 'Tischenko'
        q = select(p.first_name + f"{' ' + x}" for p in Person if p.id == 1)
        self.assertEqual(set(q), {'Alexander Tischenko'})

        q2 = select("""p.first_name + f"{' ' + x}" for p in Person if p.id == 1""")
        self.assertEqual(set(q2), {'Alexander Tischenko'})



import textwrap
import unittest
import ast
import sys

from foalorm.orm.decompiling import Decompiler
from foalorm.orm.asttranslation import ast2src


def generate_gens():
    patterns = [
        '(x * y) * [z * j)',
        '([x * y) * z) * j',
        '(x * [y * z)) * j',
        'x * ([y * z) * j)',
        'x * (y * [z * j))'
    ]

    ops = ('and', 'or')
    nots = (True, False)

    result = []

    for pat in patterns:
        cur = pat
        for op1 in ops:
            for op2 in ops:
                for op3 in ops:
                    res = cur.replace('*', op1, 1)
                    res = res.replace('*', op2, 1)
                    res = res.replace('*', op3, 1)
                    result.append(res)

    final = []

    for res in result:
        for par1 in nots:
            for par2 in nots:
                for a in nots:
                    for b in nots:
                        for c in nots:
                            for d in nots:
                                cur = res.replace('(', 'not(') if not par1 else res
                                if not par2:
                                    cur = cur.replace('[', 'not(')
                                else:
                                    cur = cur.replace('[', '(')
                                if not a: cur = cur.replace('x', 'not x')
                                if not b: cur = cur.replace('y', 'not y')
                                if not c: cur = cur.replace('z', 'not z')
                                if not d: cur = cur.replace('j', 'not j')
                                final.append(cur)

    return final

def create_test(gen):
    def wrapped_test(self):
        def get_condition_values(cond):
            result = []
            vals = (True, False)
            for x in vals:
                for y in vals:
                    for z in vals:
                        for j in vals:
                            result.append(eval(cond, {'x': x, 'y': y, 'z': z, 'j': j}))
            return result
        src1 = '(a for a in [] if %s)' % gen
        src2 = 'lambda x, y, z, j: (%s)' % gen
        src3 = '(m for m in [] if %s for n in [] if %s)' % (gen, gen)

        code1 = compile(src1, '<?>', 'eval').co_consts[0]
        ast1 = Decompiler(code1).ast
        res1 = ast2src(ast1).replace('.0', '[]')
        res1 = res1[res1.find('if')+2:-1]

        code2 = compile(src2, '<?>', 'eval').co_consts[0]
        ast2 = Decompiler(code2).ast
        res2 = ast2src(ast2).replace('.0', '[]')
        res2 = res2[res2.find(':')+1:]

        code3 = compile(src3, '<?>', 'eval').co_consts[0]
        ast3 = Decompiler(code3).ast
        res3 = ast2src(ast3).replace('.0', '[]')
        res3 = res3[res3.find('if')+2: res3.rfind('for')-1]

        if get_condition_values(gen) != get_condition_values(res1):
            self.fail("Incorrect generator decompilation: %s -> %s" % (gen, res1))

        if get_condition_values(gen) != get_condition_values(res2):
            self.fail("Incorrect lambda decompilation: %s -> %s" % (gen, res2))

        if get_condition_values(gen) != get_condition_values(res3):
            self.fail("Incorrect multi-for generator decompilation: %s -> %s" % (gen, res3))

    return wrapped_test


class TestDecompiler(unittest.TestCase):
    def assertDecompilesTo(self, src, expected):
        # skip test due to ast.dump has no indent parameter
        if sys.version_info[:2] <= (3, 8):
            return

        code = compile(src, '<?>', 'eval').co_consts[0]
        import dis
        print(dis.dis(code))
        dc = Decompiler(code)
        expected = textwrap.dedent(expected).strip()
        self.maxDiff = None
        self.assertMultiLineEqual(self._strip_each_line(expected), ast.dump(dc.ast))

    def _strip_each_line(self, string: str) -> str:
        return ' '.join(
            l.strip() for l in string.splitlines()
        ).replace('( ', '(').replace('[ ', '[').strip()

    def test_ast1(self):
        self.assertDecompilesTo(
            '(a for a in [] if x and y and z and j)',
            """
            GeneratorExp(
              elt=Name(id='a', ctx=Load()),
              generators=[
                comprehension(
                  target=Name(id='a', ctx=Store()),
                  iter=Name(id='.0', ctx=Load()),
                  ifs=[
                    BoolOp(
                      op=And(),
                      values=[
                        Name(id='x', ctx=Load()),
                        Name(id='y', ctx=Load()),
                        Name(id='z', ctx=Load()),
                        Name(id='j', ctx=Load())])],
                  is_async=0)])
            """)

    def test_ast2(self):
        self.assertDecompilesTo(
            'lambda x, y, z, j: (x and y and z and j)',
            """
            BoolOp(
              op=And(),
              values=[
                Name(id='x', ctx=Load()),
                Name(id='y', ctx=Load()),
                Name(id='z', ctx=Load()),
                Name(id='j', ctx=Load())])
            """)

    def test_ast3(self):
        self.assertDecompilesTo(
            '(m for m in [] if x and y and z and j for n in [] if x and y and z and j)',
            """
            GeneratorExp(
              elt=Name(id='m', ctx=Load()),
              generators=[
                comprehension(
                  target=Name(id='m', ctx=Store()),
                  iter=Name(id='.0', ctx=Load()),
                  ifs=[
                    BoolOp(
                      op=And(),
                      values=[
                        Name(id='x', ctx=Load()),
                        Name(id='y', ctx=Load()),
                        Name(id='z', ctx=Load()),
                        Name(id='j', ctx=Load())])],
                  is_async=0),
                comprehension(
                  target=Name(id='n', ctx=Store()),
                  iter=Constant(value=()),
                  ifs=[
                    BoolOp(
                      op=And(),
                      values=[
                        Name(id='x', ctx=Load()),
                        Name(id='y', ctx=Load()),
                        Name(id='z', ctx=Load()),
                        Name(id='j', ctx=Load())])],
                  is_async=0)])
            """)


for i, gen in enumerate(generate_gens()):
    test_method = create_test(gen)
    test_method.__name__ = 'test_decompiler_%d' % i
    setattr(TestDecompiler, test_method.__name__, test_method)



from __future__ import absolute_import, print_function, division
from foalorm.py23compat import PY38

import ast

from functools import update_wrapper

from foalorm.utils import HashableDict, throw, copy_ast

class TranslationError(Exception): pass

pre_method_caches = {}
post_method_caches = {}

class ASTTranslator(object):
    def __init__(translator, tree):
        translator.tree = tree
        translator_cls = translator.__class__
        pre_method_caches.setdefault(translator_cls, {})
        post_method_caches.setdefault(translator_cls, {})
    def dispatch(translator, node):
        translator_cls = translator.__class__
        pre_methods = pre_method_caches[translator_cls]
        post_methods = post_method_caches[translator_cls]
        if isinstance(node, (ast.BoolOp, ast.BinOp, ast.UnaryOp)):
            node_cls = node.op.__class__
        else:
            node_cls = node.__class__

        try:
            pre_method = pre_methods[node_cls]
        except KeyError:
            pre_method = getattr(translator_cls, 'pre' + node_cls.__name__, translator_cls.default_pre)
            pre_methods[node_cls] = pre_method

        stop = translator.call(pre_method, node)
        if stop: return

        for child in get_child_nodes(node):
            translator.dispatch(child)

        try:
            post_method = post_methods[node_cls]
        except KeyError:
            post_method = getattr(translator_cls, 'post' + node_cls.__name__, translator_cls.default_post)
            post_methods[node_cls] = post_method
        translator.call(post_method, node)
    def call(translator, method, node):
        return method(translator, node)
    def default_pre(translator, node):
        pass
    def default_post(translator, node):
        pass

def priority(p):
    def decorator(func):
        def new_func(translator, node):
            node.priority = p
            for child in get_child_nodes(node):
                if getattr(child, 'priority', 0) >= p:
                    child.src = '(%s)' % child.src
            return func(translator, node)
        return update_wrapper(new_func, func)
    return decorator


def binop_src(op, node):
    return op.join((node.left.src, node.right.src))


def ast2src(tree):
    src = getattr(tree, 'src', None)
    if src is not None:
        return src
    PythonTranslator(tree)
    return tree.src


def get_child_nodes(node):
    for child in ast.iter_child_nodes(node):
        if not isinstance(child, (ast.expr_context, ast.boolop, ast.unaryop, ast.operator)):
            yield child


class PythonTranslator(ASTTranslator):
    def __init__(translator, tree):
        ASTTranslator.__init__(translator, tree)
        translator.top_level_f_str = None
        translator.dispatch(tree)
    def call(translator, method, node):
        node.src = method(translator, node)
    def default_pre(translator, node):
        if getattr(node, 'src', None) is not None:
            return True  # node.src is already calculated, stop dispatching
    def default_post(translator, node):
        throw(NotImplementedError, node)
    def postGeneratorExp(translator, node):
        return '(' + node.elt.src + ' ' + ' '.join(gen.src for gen in node.generators) + ')'
    def postcomprehension(translator, node):
        src = 'for %s in %s' % (node.target.src, node.iter.src)
        if node.ifs:
            ifs = ' '.join('if ' + if_.src for if_ in node.ifs)
            src += ' ' + ifs
        return src
    def postGenExprIf(translator, node):
        return 'if %s' % node.test.src
    def postExpr(translator, node):
        return node.value.src
    def postIfExp(translator, node):
        return '%s if %s else %s' % (node.body.src, node.test.src, node.orelse.src)
    def postLambda(translator, node):
        return 'lambda %s: %s' % (node.args.src, node.body.src)
    def postarguments(translator, node):
        if node.defaults:
            nodef_args = node.args[:-len(node.defaults)]
            def_args = node.args[-len(node.defaults):]
        else:
            nodef_args = node.args
            def_args = []

        result = [arg.arg for arg in nodef_args]
        result.extend('%s=%s' % (arg.arg, default.src) for arg, default in zip(def_args, node.defaults))
        if node.vararg:
            result.append('*%s' % node.vararg.arg)
        if node.kwarg:
            result.append('**%s' % node.kwarg.arg)
        return ', '.join(result)
    def postarg(translator, node):
        return node.arg
    @priority(14)
    def postOr(translator, node):
        return ' or '.join(expr.src for expr in node.values)
    @priority(13)
    def postAnd(translator, node):
        return ' and '.join(expr.src for expr in node.values)
    @priority(12)
    def postNot(translator, node):
        return 'not ' + node.operand.src
    @priority(11)
    def postCompare(translator, node):
        result = [ node.left.src ]
        for op, expr in zip(node.ops, node.comparators):
            result.extend((op.src, expr.src))
        return ' '.join(result)
    def postEq(translator, node):
        return '=='
    def postNotEq(translator, node):
        return '!='
    def postLt(translator, node):
        return '<'
    def postLtE(translator, node):
        return '<='
    def postGt(translator, node):
        return '>'
    def postGtE(translator, node):
        return '>='
    def postIs(translator, node):
        return 'is'
    def postIsNot(translator, node):
        return 'is not'
    def postIn(translator, node):
        return 'in'
    def postNotIn(translator, node):
        return 'not in'
    @priority(10)
    def postBitOr(translator, node):
        return ' | '.join((node.left.src, node.right.src))
    @priority(9)
    def postBitXor(translator, node):
        return ' ^ '.join((node.left.src, node.right.src))
    @priority(8)
    def postBitAnd(translator, node):
        return ' & '.join((node.left.src, node.right.src))
    @priority(7)
    def postLShift(translator, node):
        return ' << '.join((node.left.src, node.right.src))
    @priority(7)
    def postRShift(translator, node):
        return ' >> '.join((node.left.src, node.right.src))
    @priority(6)
    def postAdd(translator, node):
        return ' + '.join((node.left.src, node.right.src))
    @priority(6)
    def postSub(translator, node):
        return ' - '.join((node.left.src, node.right.src))
    @priority(5)
    def postMult(translator, node):
        return ' * '.join((node.left.src, node.right.src))
    @priority(5)
    def postMatMult(translator, node):
        throw(NotImplementedError)
    @priority(5)
    def postDiv(translator, node):
        return ' / '.join((node.left.src, node.right.src))
    @priority(5)
    def postFloorDiv(translator, node):
        return ' // '.join((node.left.src, node.right.src))
    @priority(5)
    def postMod(translator, node):
        return ' % '.join((node.left.src, node.right.src))
    @priority(4)
    def postUSub(translator, node):
        return '-' + node.operand.src
    @priority(4)
    def postUAdd(translator, node):
        return '+' + node.operand.src
    @priority(4)
    def postInvert(translator, node):
        return '~' + node.expr.src
    @priority(3)
    def postPow(translator, node):
        return binop_src(' ** ', node)
    def postAttribute(translator, node):
        node.priority = 2
        return '.'.join((node.value.src, node.attr))
    def postCall(translator, node):
        node.priority = 2
        if len(node.args) == 1 and isinstance(node.args[0], ast.GeneratorExp):
            return node.func.src + node.args[0].src
        args = [ arg.src for arg in node.args ] + [ kw.src for kw in node.keywords ]
        return '%s(%s)' % (node.func.src, ', '.join(args))
    def postkeyword(translator, node):
        if node.arg is None:
            return '**' + node.value.src
        return '%s=%s' % (node.arg, node.value.src)
    def postStarred(translator, node):
        return '*' + node.value.src
    def postSubscript(translator, node):
        node.priority = 2
        x = node.slice
        if isinstance(x, ast.Index):
            x = x.value
        if isinstance(x, ast.Tuple):
            key = ', '.join([elt.src for elt in x.elts])
        elif isinstance(x, ast.Constant) and isinstance(x.value, tuple):
            key = repr(x.value)[1:-1]
        else:
            key = x.src
        return '%s[%s]' % (node.value.src, key)
    def postIndex(translator, node):  # Python <= 3.7
        return node.value.src
    def postSlice(translator, node):
        result = []
        if node.lower:
            result.append(node.lower.src)
        result.append(':')
        if node.upper:
            result.append(node.upper.src)
        if node.step:
            result.append(':')
            result.append(node.step.src)
        return ''.join(result)
    def postConstant(translator, node):
        node.priority = 1
        value = node.value
        if type(value) is float: # for Python < 2.7
            s = str(value)
            if float(s) == value: return s
        return repr(value)
    def postNameConstant(translator, node):  # Python <= 3.7
        return repr(node.value)
    def postNum(translator, node):  # Python <= 3.7
        node.priority = 1
        return repr(node.n)
    def postStr(translator, node):  # Python <= 3.7
        node.priority = 1
        return repr(node.s)
    def postBytes(translator, node):  # Python <= 3.7
        node.priority = 1
        return repr(node.s)
    def postList(translator, node):
        node.priority = 1
        return '[%s]' % ', '.join(item.src for item in node.elts)
    def postTuple(translator, node):
        node.priority = 1
        if len(node.elts) == 1:
            return '(%s,)' % node.elts[0].src
        return '(%s)' % ', '.join(elt.src for elt in node.elts)
    def postDict(translator, node):
        node.priority = 1
        return '{%s}' % ', '.join('%s:%s' % (key.src, value.src) for key, value in zip(node.keys, node.values))
    def postSet(translator, node):
        node.priority = 1
        return '{%s}' % ', '.join(item.src for item in node.nodes)
    def postName(translator, node):
        node.priority = 1
        return node.id
    def postJoinedStr(self, node):
        result = []
        for item in node.values:
            if isinstance(item, ast.Constant):
                assert isinstance(item.value, str)
                result.append(item.value)
            elif not PY38 and isinstance(item, ast.Str):  # Python 3.7
                result.append(item.s)
            elif isinstance(item, ast.FormattedValue):
                if item.conversion == -1:
                    src = '{%s}' % item.value.src
                else:
                    src = '{%s!%s}' % (item.value.src, chr(item.conversion))
                result.append(src)
            else:
                assert False
        return "f%r" % ''.join(result)
    def postFormattedValue(self, node):
        return node.value.src


nonexternalizable_types = (ast.keyword, ast.Starred, ast.Slice, ast.List, ast.Tuple)


class PreTranslator(ASTTranslator):
    def __init__(translator, tree, globals, locals,
                 special_functions, const_functions, outer_names=()):
        ASTTranslator.__init__(translator, tree)
        translator.globals = globals
        translator.locals = locals
        translator.special_functions = special_functions
        translator.const_functions = const_functions
        translator.contexts = []
        if outer_names:
            translator.contexts.append(outer_names)
        translator.externals = externals = set()
        translator.dispatch(tree)
        for node in externals.copy():
            if isinstance(node, nonexternalizable_types) or node.constant and not isinstance(node, ast.Constant):
                node.external = False
                externals.remove(node)
                externals.update(node for node in get_child_nodes(node) if node.external and not node.constant)
    def dispatch(translator, node):
        node.external = node.constant = None
        ASTTranslator.dispatch(translator, node)
        children = list(get_child_nodes(node))
        if node.external is None and children and all(
                getattr(child, 'external', False) and not getattr(child, 'raw_sql', False) for child in children):
            node.external = True
        if node.external and not node.constant:
            externals = translator.externals
            externals.difference_update(children)
            externals.add(node)
    def preGeneratorExp(translator, node):
        translator.contexts.append(set())
        dispatch = translator.dispatch
        for i, qual in enumerate(node.generators):
            dispatch(qual.iter)
            dispatch(qual.target)
            for if_ in qual.ifs:
                dispatch(if_)
        dispatch(node.elt)
        translator.contexts.pop()
        return True
    def preLambda(translator, node):
        #if node.varargs or node.kwargs or node.defaults:
        #    throw(NotImplementedError)
        context = set(arg.arg for arg in node.args.args)
        if node.args.vararg:
            context.add(node.args.vararg.arg)
        if node.args.kwarg:
            context.add(node.args.kwarg.arg)
        translator.contexts.append(context)
        translator.dispatch(node.body)
        translator.contexts.pop()
        return True
    def postName(translator, node):
        name = node.id
        if isinstance(node.ctx, ast.Store):
            if name.startswith('__'):
                throw(TranslationError, 'Illegal name: %r' % name)
            translator.contexts[-1].add(name)
            return
        elif isinstance(node.ctx, ast.Load):
            for context in translator.contexts:
                if name in context:
                    return
            node.external = True
        else:
            assert False, type(node.ctx)
    def postSlice(translator, node):
        if node.lower is None and node.upper is None and node.step is None:
            node.external = node.constant = True
    def postStarred(translator, node):
        node.external = True
    def postConstant(translator, node):
        node.external = node.constant = True
    def postNum(translator, node):  # Python <= 3.7
        node.external = node.constant = True
    def postStr(translator, node):  # Python <= 3.7
        node.external = node.constant = True
    def postBytes(translator, node):  # Python <= 3.7
        node.external = node.constant = True
    def postDict(translator, node):
        node.external = True
    def postList(translator, node):
        node.external = True
    def postkeyword(translator, node):
        node.constant = node.value.constant
    def postIndex(translator, node):  # Python <= 3.7
        node.constant = node.value.constant
    def postCall(translator, node):
        func_node = node.func
        if not func_node.external:
            return
        attrs = []
        while isinstance(func_node, ast.Attribute):
            attrs.append(func_node.attr)
            func_node = func_node.value
        if not isinstance(func_node, ast.Name):
            return
        attrs.append(func_node.id)
        expr = '.'.join(reversed(attrs))
        x = eval(expr, translator.globals, translator.locals)
        try:
            hash(x)
        except TypeError:
            return

        if x in translator.special_functions:
            if x.__name__ == 'raw_sql':
                node.raw_sql = True
            elif x is getattr:
                attr_node = node.args[1]
                attr_node.parent_node = node
            else:
                node.external = False
        elif x in translator.const_functions:
            for arg in node.args:
                if not arg.constant:
                    return
            if any(not arg.constant for arg in node.args if isinstance(arg, ast.Starred)):
                return
            if any(not kwarg.constant for kwarg in node.keywords if kwarg.arg is None):
                return
            node.constant = True
    def postCompare(translator, node):
        for op in node.ops:
            op.external = op.constant = True
    def post_binop(translator, node):
        pass

    postBitOr = postBitXor = postBitAnd = postLShift = postRShift \
        = postAdd = postSub = postMult = postMatMult = postDiv = postFloorDiv = postMod = post_binop

extractors_cache = {}

def create_extractors(code_key, tree, globals, locals, special_functions, const_functions, outer_names=()):
    result = extractors_cache.get(code_key)
    if not result:
        pretranslator = PreTranslator(tree, globals, locals, special_functions, const_functions, outer_names)
        extractors = {}
        for node in pretranslator.externals:
            src = node.src = ast2src(node)
            if src == '.0':
                def extractor(globals, locals):
                    return locals['.0']
            else:
                filename = '<foalorm ' + src + '>'
                code = compile(src, filename, 'eval')
                def extractor(globals, locals, code=code):
                    return eval(code, globals, locals)
            extractors[src] = extractor
        result = extractors_cache[code_key] = tree, extractors
    return result



from __future__ import absolute_import, print_function, division

from bottle import HTTPResponse, HTTPError
from foalorm.orm.core import db_session

def is_allowed_exception(e):
    return isinstance(e, HTTPResponse) and not isinstance(e, HTTPError)

class FoalORMPlugin(object):
    name = 'foalorm'
    api  = 2
    def apply(self, callback, route):
        return db_session(allowed_exceptions=is_allowed_exception)(callback)



from foalorm.orm import db_session
from flask import request

def _enter_session():
    session = db_session()
    request.foalorm_session = session
    session.__enter__()

def _exit_session(exception):
    session = getattr(request, 'foalorm_session', None)
    if session is not None:
        session.__exit__(exc=exception)

class FoalORM(object):
    def __init__(self, app=None):
        self.app = None
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        self.app.before_request(_enter_session)
        self.app.teardown_request(_exit_session)